# -*- coding: utf-8 -*-

from hbshare.fe.xwq.analysis.orm.hbdb import HBDB
from datetime import datetime
from sklearn.linear_model import LinearRegression as LR
import os
import numpy as np
import pandas as pd
import statsmodels.api as sm

from WindPy import w
w.start()  # 默认命令超时时间为120秒，如需设置超时时间可以加入waitTime参数，例如waitTime=60,即设置命令超时时间为60秒
w.isconnected()  # 判断WindPy是否已经登录成功

from matplotlib.ticker import FuncFormatter
import matplotlib.pyplot as plt
import seaborn as sns
plt.rcParams['font.sans-serif'] = ['kaiti']  # 用来正常显示中文标签
plt.rcParams['axes.unicode_minus'] = False  # 用来正常显示负号
line_color_list = ['#F04950', '#6268A2', '#959595', '#333335', '#EE703F', '#7E4A9B', '#8A662C',
                  '#44488E', '#BA67E9', '#3FAEEE']
bar_color_list = ['#C94649', '#EEB2B4', '#E1777A', '#D57C56', '#E39A79', '#DB8A66', '#E5B88C',
                  '#8588B7', '#B4B6D1', '#55598D', '#628497', '#A9C6CB', '#866EA9', '#B79BC7',
                  '#7D7D7E', '#CACACA', '#A7A7A8', '#606063', '#C4C4C4', '#99999B', '#B7B7B7']
area_color_list = ['#D55659', '#E1777A', '#DB8A66', '#E5B88C', '#EEB2B4', '#D57C56', '#E39A79',
                   '#8588B7', '#626697', '#866EA9', '#B79BC7', '#B4B6D1', '#628497', '#A9C6CB',
                   '#7D7D7E', '#A7A7A8', '#99999B', '#B7B7B7', '#CACACA', '#969696', '#C4C4C4']
new_color_list = ['#F04950', '#959595', '#6268A2', '#333335', '#D57C56', '#628497']
industry_theme_dic = {'银行': '大金融', '非银金融': '大金融', '房地产': '大金融',
                      '食品饮料': '消费', '家用电器': '消费', '医药生物': '消费', '社会服务': '消费', '农林牧渔': '消费', '商贸零售': '消费', '美容护理': '消费',
                      '通信': 'TMT', '计算机': 'TMT', '电子': 'TMT', '传媒': 'TMT', '国防军工': 'TMT',
                      '交通运输': '制造', '机械设备': '制造', '汽车': '制造', '纺织服饰': '制造', '轻工制造': '制造', '电力设备': '制造',
                      '钢铁': '周期', '有色金属': '周期', '建筑装饰': '周期', '建筑材料': '周期', '基础化工': '周期', '石油石化': '周期', '煤炭': '周期', '公用事业': '周期', '环保': '周期',
                      '综合': '其他'}

def to_100percent(temp, position):
    return '%1.0f'%(temp * 100) + '%'


def get_date(start_date, end_date):
    calendar_df = HBDB().read_cal(start_date, end_date)
    calendar_df = calendar_df.rename(columns={'JYRQ': 'CALENDAR_DATE', 'SFJJ': 'IS_OPEN', 'SFZM': 'IS_WEEK_END', 'SFYM': 'IS_MONTH_END'})
    calendar_df['CALENDAR_DATE'] = calendar_df['CALENDAR_DATE'].astype(str)
    calendar_df = calendar_df.sort_values('CALENDAR_DATE')
    calendar_df['IS_OPEN'] = calendar_df['IS_OPEN'].astype(int).replace({0: 1, 1: 0})
    calendar_df['YEAR_MONTH'] = calendar_df['CALENDAR_DATE'].apply(lambda x: x[:6])
    calendar_df['MONTH'] = calendar_df['CALENDAR_DATE'].apply(lambda x: x[4:6])
    calendar_df['MONTH_DAY'] = calendar_df['CALENDAR_DATE'].apply(lambda x: x[4:])
    calendar_df = calendar_df[(calendar_df['CALENDAR_DATE'] >= start_date) & (calendar_df['CALENDAR_DATE'] <= end_date)]
    trade_df = calendar_df[calendar_df['IS_OPEN'] == 1].rename(columns={'CALENDAR_DATE': 'TRADE_DATE'})
    trade_df = trade_df[(trade_df['TRADE_DATE'] >= start_date) & (trade_df['TRADE_DATE'] <= end_date)]
    report_df = calendar_df.drop_duplicates('YEAR_MONTH', keep='last').rename(columns={'CALENDAR_DATE': 'REPORT_DATE'})
    report_df = report_df[report_df['MONTH_DAY'].isin(['0331', '0630', '0930', '1231'])]
    report_df = report_df[(report_df['REPORT_DATE'] >= start_date) & (report_df['REPORT_DATE'] <= end_date)]
    report_trade_df = calendar_df[calendar_df['IS_OPEN'] == 1].rename(columns={'CALENDAR_DATE': 'TRADE_DATE'})
    report_trade_df = report_trade_df.sort_values('TRADE_DATE').drop_duplicates('YEAR_MONTH', keep='last')
    report_trade_df = report_trade_df[report_trade_df['MONTH'].isin(['03', '06', '09', '12'])]
    report_trade_df = report_trade_df[(report_trade_df['TRADE_DATE'] >= start_date) & (report_trade_df['TRADE_DATE'] <= end_date)]
    calendar_trade_df = calendar_df[['CALENDAR_DATE']].merge(trade_df[['TRADE_DATE']], left_on=['CALENDAR_DATE'], right_on=['TRADE_DATE'], how='left')
    calendar_trade_df['TRADE_DATE'] = calendar_trade_df['TRADE_DATE'].fillna(method='ffill')
    calendar_trade_df = calendar_trade_df[(calendar_trade_df['TRADE_DATE'] >= start_date) & (calendar_trade_df['TRADE_DATE'] <= end_date)]
    return calendar_df, report_df, trade_df, report_trade_df, calendar_trade_df


def get_industry_info():
    industry_info = HBDB().read_industry_info()
    industry_info = industry_info.rename(columns={'flmc': 'INDUSTRY_NAME', 'zsdm': 'INDUSTRY_ID', 'hyhfbz': 'INDUSTRY_VERSION', 'fljb': 'INDUSTRY_TYPE', 'qsrq': 'BEGIN_DATE', 'jsrq': 'END_DATE', 'sfyx': 'IS_NEW'})
    industry_info = industry_info.dropna(subset=['BEGIN_DATE'])
    industry_info['END_DATE'] = industry_info['END_DATE'].replace('', np.nan).fillna('20990101')
    industry_info['BEGIN_DATE'] = industry_info['BEGIN_DATE'].apply(lambda x: datetime.strptime(x, '%Y/%m/%d').strftime('%Y%m%d') if '/' in x else x)
    industry_info['END_DATE'] = industry_info['END_DATE'].apply(lambda x: datetime.strptime(x, '%Y/%m/%d').strftime('%Y%m%d') if '/' in x else x)
    industry_info['BEGIN_DATE'] = industry_info['BEGIN_DATE'].astype(int).astype(str)
    industry_info['END_DATE'] = industry_info['END_DATE'].astype(int).astype(str)
    industry_info['INDUSTRY_VERSION'] = industry_info['INDUSTRY_VERSION'].astype(int)
    industry_info['INDUSTRY_TYPE'] = industry_info['INDUSTRY_TYPE'].astype(int)
    industry_info['IS_NEW'] = industry_info['IS_NEW'].astype(int)
    industry_info = industry_info[industry_info['INDUSTRY_VERSION'] == 3]
    return industry_info


def get_industry_symbol():
    industry_symbol = HBDB().read_industry_symbol()
    industry_symbol = industry_symbol.rename(columns={'hyhfbz': 'INDUSTRY_VERSION', 'fldm': 'INDUSTRY_ID', 'flmc': 'INDUSTRY_NAME', 'zsdm': 'INDEX_SYMBOL', 'fljb': 'INDUSTRY_TYPE', 'qsrq': 'BEGIN_DATE', 'jsrq': 'END_DATE', 'sfyx': 'IS_NEW'})
    industry_symbol = industry_symbol.dropna(subset=['BEGIN_DATE'])
    industry_symbol['END_DATE'] = industry_symbol['END_DATE'].replace('', np.nan).fillna('20990101')
    industry_symbol['BEGIN_DATE'] = industry_symbol['BEGIN_DATE'].astype(int).astype(str)
    industry_symbol['END_DATE'] = industry_symbol['END_DATE'].astype(int).astype(str)
    industry_symbol['INDUSTRY_VERSION'] = industry_symbol['INDUSTRY_VERSION'].astype(int)
    industry_symbol['INDUSTRY_TYPE'] = industry_symbol['INDUSTRY_TYPE'].astype(int)
    industry_symbol['IS_NEW'] = industry_symbol['IS_NEW'].astype(int)
    industry_symbol = industry_symbol[industry_symbol['INDUSTRY_VERSION'] == 3]
    return industry_symbol


class valuation_earning_contribution:
    def __init__(self, data_path, start_date, end_date, index):
        self.data_path = data_path
        self.start_date = start_date
        self.end_date = end_date
        self.start_date_hyphen = datetime.strptime(start_date, '%Y%m%d').strftime('%Y-%m-%d')
        self.end_date_hyphen = datetime.strptime(end_date, '%Y%m%d').strftime('%Y-%m-%d')
        self.index = index
        self.index_name_dict = {'000016.SH': '上证50', '000300.SH': '沪深300', '000905.SH': '中证500', '000852.SH': '中证1000', '881001.WI': '万得全A', 'HSI.HI': '恒生指数',
                                'DJR': '大金融', 'XF': '消费', 'TMT': 'TMT', 'ZZ': '制造', 'ZQ': '周期', 'QT': '其他'}
        self.industry_info = get_industry_info()
        self.industry_info = self.industry_info[self.industry_info['INDUSTRY_TYPE'] == 1]
        self.industry_info = self.industry_info[self.industry_info['IS_NEW'] == 1]
        self.industry_info = self.industry_info[['INDUSTRY_ID', 'INDUSTRY_NAME']]
        self.industry_info['INDUSTRY_ID_SI'] = self.industry_info['INDUSTRY_ID'].apply(lambda x: str(x) + '.SI')
        self.index_name_dict.update(self.industry_info.set_index('INDUSTRY_ID_SI')['INDUSTRY_NAME'].to_dict())
        self.load()

    def load(self):
        self.calendar_df, self.report_df, self.trade_df, self.report_trade_df, self.calendar_trade_df = get_date(self.start_date, self.end_date)
        self.year_df = self.report_df[self.report_df['REPORT_DATE'].str.slice(4, 6) == '12']

        if self.index not in ['DJR', 'XF', 'TMT', 'ZZ', 'ZQ', 'QT']:
            index_data_path = '{0}{1}.hdf'.format(self.data_path, self.index.split('.')[0])
            if os.path.isfile(index_data_path):
                existed_index_data = pd.read_hdf(index_data_path, key='table')
                max_date = max(existed_index_data['TRADE_DATE'])
                start_date = max(str(max_date), '20071231')
            else:
                existed_index_data = pd.DataFrame()
                start_date = '20071231'
            if start_date < self.end_date:
                start_date_hyphen = datetime.strptime(start_date, '%Y%m%d').strftime('%Y-%m-%d')
                index_data_update = w.wsd(self.index, "close,pe_ttm,eps_ttm", start_date_hyphen, self.end_date_hyphen, "Days=Alldays", usedf=True)[1].reset_index()
                index_data_update = index_data_update.rename(columns={'index': 'TRADE_DATE'})
                index_data_update['TRADE_DATE'] = index_data_update['TRADE_DATE'].apply(lambda x: str(x).replace('-', ''))
                self.index_data = pd.concat([existed_index_data, index_data_update], ignore_index=True)
                self.index_data = self.index_data.drop_duplicates()
                self.index_data.to_hdf(index_data_path, key='table', mode='w')
            self.index_data = pd.read_hdf(index_data_path, key='table')
            self.index_data['TRADE_DATE'] = self.index_data['TRADE_DATE'].astype(str)
            self.index_data = self.index_data.set_index('TRADE_DATE').sort_index()
            self.index_data = self.index_data.fillna(method='ffill')
        else:
            index_daily_k_path = '{0}index_daily_k.hdf'.format(self.data_path)
            if os.path.isfile(index_daily_k_path):
                existed_index_daily_k = pd.read_hdf(index_daily_k_path, key='table')
                max_date = max(existed_index_daily_k['jyrq'])
                start_date = max(str(max_date), '20071231')
            else:
                existed_index_daily_k = pd.DataFrame()
                start_date = '20071231'
            trade_df = self.trade_df[(self.trade_df['TRADE_DATE'] > start_date) & (self.trade_df['TRADE_DATE'] <= self.end_date)]
            index_daily_k_list = []
            for date in trade_df['TRADE_DATE'].unique().tolist():
                index_daily_k_date = HBDB().read_index_daily_k_given_date(start_date, self.industry_info['INDUSTRY_ID'].unique().tolist())
                index_daily_k_list.append(index_daily_k_date)
                print(date)
            self.index_daily_k = pd.concat([existed_index_daily_k] + index_daily_k_list, ignore_index=True)
            self.index_daily_k.to_hdf(index_daily_k_path, key='table', mode='w')
            self.index_daily_k = pd.read_hdf(index_daily_k_path, key='table')
            self.index_daily_k = self.index_daily_k[['zqdm', 'jyrq', 'ltsz', 'spjg', 'pe']]
            self.index_daily_k.columns = ['INDUSTRY_ID', 'TRADE_DATE', 'NEG_MARKET_VALUE', 'CLOSE_INDEX', 'PE']

            self.theme_industry = pd.DataFrame(list(industry_theme_dic.items()))
            self.theme_industry.columns = ['INDUSTRY_NAME', 'THEME_NAME']
            self.theme_industry = self.theme_industry.merge(self.industry_info[['INDUSTRY_NAME', 'INDUSTRY_ID']], on=['INDUSTRY_NAME'], how='left')
            needed_industry_list = self.theme_industry[self.theme_industry['THEME_NAME'] == self.index_name_dict[self.index]]['INDUSTRY_ID'].unique().tolist()
            theme_nmv = self.index_daily_k[self.index_daily_k['INDUSTRY_ID'].isin(needed_industry_list)].pivot(index='TRADE_DATE', columns='INDUSTRY_ID', values='NEG_MARKET_VALUE').sort_index()
            theme_nmv = theme_nmv.apply(lambda x: x / x.sum(), axis=1)
            theme_ret = self.index_daily_k[self.index_daily_k['INDUSTRY_ID'].isin(needed_industry_list)].pivot(index='TRADE_DATE', columns='INDUSTRY_ID', values='CLOSE_INDEX').sort_index()
            theme_ret = theme_ret.replace(0.0, np.nan).fillna(method='ffill').pct_change()
            theme_pe = self.index_daily_k[self.index_daily_k['INDUSTRY_ID'].isin(needed_industry_list)].pivot(index='TRADE_DATE', columns='INDUSTRY_ID', values='PE').sort_index()
            theme_ret = (theme_nmv * theme_ret).sum(axis=1)
            theme_nav = pd.DataFrame((theme_ret + 1).cumprod(), columns=['CLOSE'])
            theme_pe = pd.DataFrame((theme_nmv * theme_pe).sum(axis=1), columns=['PE_TTM'])
            self.index_data = pd.concat([theme_nav, theme_pe], axis=1)
            self.index_data = self.index_data.reset_index().rename(columns={'index': 'TRADE_DATE'})
            self.index_data['TRADE_DATE'] = self.index_data['TRADE_DATE'].astype(str)
            self.index_data = self.index_data.set_index('TRADE_DATE').sort_index()
            self.index_data = self.index_data.fillna(method='ffill')
        return

    def get_result_1(self):
        # P =  PE * EPS, PE取PE_TTM, EPS取EPS_TTM
        # 指数涨跌幅 = PE涨跌幅 + EPS涨跌幅 (+ 不能被PE\EPS涨跌幅解释部分)
        index_data = self.index_data[self.index_data.index.isin(self.year_df['REPORT_DATE'].unique().tolist())]
        index_data_change = index_data.pct_change().dropna(how='all', axis=1).dropna()

        index_data_change['IS_ADD'] = index_data_change.apply(lambda x: 1 if np.sign(x['PE_TTM']) * np.sign(x['EPS_TTM']) == 1 else 0, axis=1)
        index_data_change['PE_TTM_DISP'] = index_data_change['PE_TTM']
        index_data_change['EPS_TTM_DISP'] = index_data_change.apply(lambda x: x['PE_TTM'] + x['EPS_TTM'] if x['IS_ADD'] == 1 else x['EPS_TTM'], axis=1)

        fig, ax = plt.subplots(figsize=(12, 6))
        bar_width = 0.3
        ax.bar(np.arange(len(index_data_change)), index_data_change['EPS_TTM_DISP'].values, bar_width, label='EPS_TTM涨跌幅', color=bar_color_list[14])
        ax.bar(np.arange(len(index_data_change)), index_data_change['PE_TTM_DISP'].values, bar_width, label='PE_TTM涨跌幅', color=bar_color_list[0])
        ax.plot(np.arange(len(index_data_change)), index_data_change['CLOSE'].values, label='{0}指数涨跌幅'.format(self.index_name_dict[self.index]),  color=bar_color_list[7])
        plt.legend(loc=8, bbox_to_anchor=(0.5, -0.25), ncol=3)
        ax.set_xticks(np.arange(len(index_data_change)))
        ax.set_xticklabels(labels=[i[:4]for i in index_data_change.index.tolist()])
        ax.yaxis.set_major_formatter(FuncFormatter(to_100percent))
        plt.title('{0}估值和盈利增长贡献拆解'.format(self.index_name_dict[self.index]), fontdict={'font': 'kaiti', 'weight': 'bold', 'size': 16})
        plt.tight_layout()
        sns.despine(top=True, right=True, left=False, bottom=False)
        plt.savefig('{0}{1}_相加.png'.format(self.data_path, self.index_name_dict[self.index]))
        return

    def get_result_2(self):
        # P =  PE * EPS, PE取PE_TTM, EPS取EPS_TTM
        # 指数涨跌幅与PE涨跌幅和EPS涨跌幅相关，做回归
        # 指数涨跌幅 = PE涨跌幅贡献 + EPS涨跌幅贡献
        index_data = self.index_data[self.index_data.index.isin(self.year_df['REPORT_DATE'].unique().tolist())]
        index_data_change = index_data.pct_change().dropna(how='all', axis=1).dropna()

        fig, ax = plt.subplots(1, 2, figsize=(12, 6))
        ax[0].scatter(index_data_change['PE_TTM'].values, index_data_change['CLOSE'].values, color=bar_color_list[0])
        ax[1].scatter(index_data_change['EPS_TTM'].values, index_data_change['CLOSE'].values, color=bar_color_list[0])
        ax[0].set_xlabel('PE涨跌幅')
        ax[0].set_ylabel('指数涨跌幅')
        ax[1].set_xlabel('EPS涨跌幅')
        ax[1].set_ylabel('指数涨跌幅')
        ax[0].xaxis.set_major_formatter(FuncFormatter(to_100percent))
        ax[0].yaxis.set_major_formatter(FuncFormatter(to_100percent))
        ax[1].xaxis.set_major_formatter(FuncFormatter(to_100percent))
        ax[1].yaxis.set_major_formatter(FuncFormatter(to_100percent))
        ax[0].set_title('{0}指数涨跌幅与估值涨跌幅关系'.format(self.index_name_dict[self.index]), fontdict={'font': 'kaiti', 'weight': 'bold', 'size': 12})
        ax[1].set_title('{0}指数涨跌幅与盈利涨跌幅关系'.format(self.index_name_dict[self.index]), fontdict={'font': 'kaiti', 'weight': 'bold', 'size': 12})
        plt.tight_layout()
        sns.despine(top=True, right=True, left=False, bottom=False)
        plt.savefig('{0}{1}_散点图.png'.format(self.data_path, self.index_name_dict[self.index]))

        model = sm.OLS(index_data_change[['CLOSE']].values, index_data_change[['PE_TTM', 'EPS_TTM']].values)
        result = model.fit()
        print(result.summary())
        coef1 = result.params[0]
        coef2 = result.params[1]

        index_data_change['PE_TTM_C'] = index_data_change['PE_TTM'] * coef1
        index_data_change['EPS_TTM_C'] = index_data_change['EPS_TTM'] * coef2
        index_data_change['IS_ADD'] = index_data_change.apply(lambda x: 1 if np.sign(x['PE_TTM_C']) * np.sign(x['EPS_TTM_C']) == 1 else 0, axis=1)
        index_data_change['PE_TTM_DISP'] = index_data_change['PE_TTM_C']
        index_data_change['EPS_TTM_DISP'] = index_data_change.apply(lambda x: x['PE_TTM_C'] + x['EPS_TTM_C'] if x['IS_ADD'] == 1 else x['EPS_TTM_C'], axis=1)

        fig, ax = plt.subplots(figsize=(12, 6))
        bar_width = 0.3
        ax.bar(np.arange(len(index_data_change)), index_data_change['EPS_TTM_DISP'].values, bar_width, label='EPS_TTM贡献', color=bar_color_list[14])
        ax.bar(np.arange(len(index_data_change)), index_data_change['PE_TTM_DISP'].values, bar_width, label='PE_TTM贡献', color=bar_color_list[0])
        ax.plot(np.arange(len(index_data_change)), index_data_change['CLOSE'].values, label='{0}指数涨跌幅'.format(self.index_name_dict[self.index]), color=bar_color_list[7])
        plt.legend(loc=8, bbox_to_anchor=(0.5, -0.25), ncol=3)
        ax.set_xticks(np.arange(len(index_data_change)))
        ax.set_xticklabels(labels=[i[:4]for i in index_data_change.index.tolist()])
        ax.yaxis.set_major_formatter(FuncFormatter(to_100percent))
        plt.title('{0}估值和盈利增长贡献拆解'.format(self.index_name_dict[self.index]), fontdict={'font': 'kaiti', 'weight': 'bold', 'size': 16})
        plt.tight_layout()
        sns.despine(top=True, right=True, left=False, bottom=False)
        plt.savefig('{0}{1}_回归2变量.png'.format(self.data_path, self.index_name_dict[self.index]))
        return

    def get_result_3(self):
        # P =  PE * EPS, PE取PE_TTM, EPS取EPS_TTM
        # 指数涨跌幅与PE涨跌幅相关，做回归，残差项作为EPS涨跌幅贡献
        # 指数涨跌幅 = PE涨跌幅贡献 + EPS涨跌幅贡献
        index_data = self.index_data[self.index_data.index.isin(self.year_df['REPORT_DATE'].unique().tolist())]
        index_data_change = index_data.pct_change().dropna(how='all', axis=1).dropna()

        model = sm.OLS(index_data_change[['CLOSE']].values, index_data_change[['PE_TTM']].values)
        result = model.fit()
        print(result.summary())
        coef1 = result.params[0]

        index_data_change['PE_TTM_C'] = index_data_change['PE_TTM'] * coef1
        index_data_change['EPS_TTM_C'] = index_data_change['CLOSE'] - index_data_change['PE_TTM_C']
        index_data_change['IS_ADD'] = index_data_change.apply(lambda x: 1 if np.sign(x['PE_TTM_C']) * np.sign(x['EPS_TTM_C']) == 1 else 0, axis=1)
        index_data_change['PE_TTM_DISP'] = index_data_change['PE_TTM_C']
        index_data_change['EPS_TTM_DISP'] = index_data_change.apply(lambda x: x['PE_TTM_C'] + x['EPS_TTM_C'] if x['IS_ADD'] == 1 else x['EPS_TTM_C'], axis=1)

        fig, ax = plt.subplots(figsize=(12, 6))
        bar_width = 0.3
        ax.bar(np.arange(len(index_data_change)), index_data_change['EPS_TTM_DISP'].values, bar_width, label='EPS_TTM贡献', color=bar_color_list[14])
        ax.bar(np.arange(len(index_data_change)), index_data_change['PE_TTM_DISP'].values, bar_width, label='PE_TTM贡献', color=bar_color_list[0])
        ax.plot(np.arange(len(index_data_change)), index_data_change['CLOSE'].values, label='{0}指数涨跌幅'.format(self.index_name_dict[self.index]), color=bar_color_list[7])
        plt.legend(loc=8, bbox_to_anchor=(0.5, -0.25), ncol=3)
        ax.set_xticks(np.arange(len(index_data_change)))
        ax.set_xticklabels(labels=[i[:4]for i in index_data_change.index.tolist()])
        ax.yaxis.set_major_formatter(FuncFormatter(to_100percent))
        plt.title('{0}估值和盈利增长贡献拆解'.format(self.index_name_dict[self.index]), fontdict={'font': 'kaiti', 'weight': 'bold', 'size': 16})
        plt.tight_layout()
        sns.despine(top=True, right=True, left=False, bottom=False)
        plt.savefig('{0}{1}_回归1变量.png'.format(self.data_path, self.index_name_dict[self.index]))
        return

if __name__ == '__main__':
   data_path = 'D:/Git/hbshare/hbshare/fe/xwq/data/valuation_earning_contribution/'
   start_date = '20071231'
   end_date = '20221231'
   industry_info = get_industry_info()
   industry_info = industry_info[industry_info['INDUSTRY_TYPE'] == 1]
   industry_info = industry_info[industry_info['IS_NEW'] == 1]
   industry_info = industry_info[['INDUSTRY_ID', 'INDUSTRY_NAME']]
   industry_list = industry_info['INDUSTRY_ID'].unique().tolist()
   industry_list = [industry + '.SI' for industry in industry_list]
   theme_list = ['DJR', 'XF', 'TMT', 'ZZ', 'ZQ', 'QT']
   index_list = ['000016.SH', '000300.SH', '000905.SH', '000852.SH', '881001.WI', 'HSI.HI'] + industry_list + theme_list
   for index in index_list:
       print(index)
       contribution = valuation_earning_contribution(data_path, start_date, end_date, index)
       if index not in ['HSI.HI'] + theme_list:
           contribution.get_result_1()
           contribution.get_result_2()
           contribution.get_result_3()
       else:
           contribution.get_result_3()


# -*- coding: utf-8 -*-

from hbshare.fe.xwq.analysis.orm.fedb import FEDB
from hbshare.fe.xwq.analysis.orm.hbdb import HBDB
from datetime import datetime, timedelta
import numpy as np
import pandas as pd


def get_date(start_date, end_date):
    calendar_df = HBDB().read_cal(start_date, end_date)
    calendar_df = calendar_df.rename(columns={'JYRQ': 'CALENDAR_DATE', 'SFJJ': 'IS_OPEN', 'SFZM': 'IS_WEEK_END', 'SFYM': 'IS_MONTH_END'})
    calendar_df['CALENDAR_DATE'] = calendar_df['CALENDAR_DATE'].astype(str)
    calendar_df = calendar_df.sort_values('CALENDAR_DATE')
    calendar_df['IS_OPEN'] = calendar_df['IS_OPEN'].astype(int).replace({0: 1, 1: 0})
    calendar_df['YEAR_MONTH'] = calendar_df['CALENDAR_DATE'].apply(lambda x: x[:6])
    calendar_df['MONTH'] = calendar_df['CALENDAR_DATE'].apply(lambda x: x[4:6])
    calendar_df['MONTH_DAY'] = calendar_df['CALENDAR_DATE'].apply(lambda x: x[4:])
    calendar_df = calendar_df[(calendar_df['CALENDAR_DATE'] >= start_date) & (calendar_df['CALENDAR_DATE'] <= end_date)]
    trade_df = calendar_df[calendar_df['IS_OPEN'] == 1].rename(columns={'CALENDAR_DATE': 'TRADE_DATE'})
    trade_df = trade_df[(trade_df['TRADE_DATE'] >= start_date) & (trade_df['TRADE_DATE'] <= end_date)]
    report_df = calendar_df.drop_duplicates('YEAR_MONTH', keep='last').rename(columns={'CALENDAR_DATE': 'REPORT_DATE'})
    report_df = report_df[report_df['MONTH_DAY'].isin(['0331', '0630', '0930', '1231'])]
    report_df = report_df[(report_df['REPORT_DATE'] >= start_date) & (report_df['REPORT_DATE'] <= end_date)]
    report_trade_df = calendar_df[calendar_df['IS_OPEN'] == 1].rename(columns={'CALENDAR_DATE': 'TRADE_DATE'})
    report_trade_df = report_trade_df.sort_values('TRADE_DATE').drop_duplicates('YEAR_MONTH', keep='last')
    report_trade_df = report_trade_df[report_trade_df['MONTH'].isin(['03', '06', '09', '12'])]
    report_trade_df = report_trade_df[(report_trade_df['TRADE_DATE'] >= start_date) & (report_trade_df['TRADE_DATE'] <= end_date)]
    calendar_trade_df = calendar_df[['CALENDAR_DATE']].merge(trade_df[['TRADE_DATE']], left_on=['CALENDAR_DATE'], right_on=['TRADE_DATE'], how='left')
    calendar_trade_df['TRADE_DATE'] = calendar_trade_df['TRADE_DATE'].fillna(method='ffill')
    calendar_trade_df = calendar_trade_df[(calendar_trade_df['TRADE_DATE'] >= start_date) & (calendar_trade_df['TRADE_DATE'] <= end_date)]
    return calendar_df, report_df, trade_df, report_trade_df, calendar_trade_df

def get_st(start_date, end_date):
    calendar_df, report_df, trade_df, report_trade_df, calendar_trade_df = get_date(start_date, end_date)
    in_st = pd.read_excel('{0}实施st.xlsx'.format(data_path))
    in_st = in_st[['代码', '实施日期']]
    in_st.columns = ['TICKER_SYMBOL', 'INTO_DATE']
    in_st['TICKER_SYMBOL'] = in_st['TICKER_SYMBOL'].apply(lambda x: x.split('.')[0])
    in_st['INTO_DATE'] = in_st['INTO_DATE'].apply(lambda x: str(x.date()).replace('-', ''))
    in_st['INTO_ST'] = 1
    in_st = in_st.drop_duplicates()
    in_st = in_st.pivot(index='INTO_DATE', columns='TICKER_SYMBOL', values='INTO_ST')
    in_st = trade_df[['TRADE_DATE']].set_index('TRADE_DATE').sort_index().merge(in_st, left_index=True, right_index=True, how='left')
    in_st = in_st.fillna(0.0)
    out_st = pd.read_excel('{0}撤销st.xlsx'.format(data_path))
    out_st = out_st[['代码', '撤销日期']]
    out_st.columns = ['TICKER_SYMBOL', 'OUT_DATE']
    out_st['TICKER_SYMBOL'] = out_st['TICKER_SYMBOL'].apply(lambda x: x.split('.')[0])
    out_st['OUT_DATE'] = out_st['OUT_DATE'].apply(lambda x: str(x.date()).replace('-', ''))
    out_st['OUT_ST'] = -1
    out_st = out_st.drop_duplicates()
    out_st = out_st.pivot(index='OUT_DATE', columns='TICKER_SYMBOL', values='OUT_ST')
    out_st = trade_df[['TRADE_DATE']].set_index('TRADE_DATE').sort_index().merge(out_st, left_index=True, right_index=True, how='left')
    out_st = out_st.fillna(0.0)
    all_tickrs = sorted(list(set(list(in_st.columns) + list(out_st.columns))))
    st = in_st.T.reindex(all_tickrs).T.fillna(0.0) + out_st.T.reindex(all_tickrs).T.fillna(0.0)
    st = st.replace(0.0, np.nan)
    st = st.fillna(method='ffill')
    st = st.replace(-1, np.nan)
    return st

def get_stock_info():
    stock_info = HBDB().read_stock_info()
    stock_info = stock_info.rename(columns={'zqdm': 'TICKER_SYMBOL', 'zqjc': 'SEC_SHORT_NAME', 'ssrq': 'ESTABLISH_DATE'})
    stock_info['ESTABLISH_DATE'] = stock_info['ESTABLISH_DATE'].dropna()
    stock_info['ESTABLISH_DATE'] = stock_info['ESTABLISH_DATE'].astype(int).astype(str)
    stock_info = stock_info.loc[stock_info['TICKER_SYMBOL'].str.len() == 6]
    stock_info = stock_info.loc[stock_info['TICKER_SYMBOL'].astype(str).str.slice(0, 1).isin(['0', '3', '6'])]
    return stock_info

def get_industry_info():
    industry_info = HBDB().read_industry_info()
    industry_info = industry_info.rename(columns={'flmc': 'INDUSTRY_NAME', 'zsdm': 'INDUSTRY_ID', 'hyhfbz': 'INDUSTRY_VERSION', 'fljb': 'INDUSTRY_TYPE', 'qsrq': 'BEGIN_DATE', 'jsrq': 'END_DATE', 'sfyx': 'IS_NEW'})
    industry_info = industry_info.dropna(subset=['BEGIN_DATE'])
    industry_info['END_DATE'] = industry_info['END_DATE'].replace('', np.nan).fillna('20990101')
    industry_info['BEGIN_DATE'] = industry_info['BEGIN_DATE'].apply(lambda x: datetime.strptime(x, '%Y/%m/%d').strftime('%Y%m%d') if '/' in x else x)
    industry_info['END_DATE'] = industry_info['END_DATE'].apply(lambda x: datetime.strptime(x, '%Y/%m/%d').strftime('%Y%m%d') if '/' in x else x)
    industry_info['BEGIN_DATE'] = industry_info['BEGIN_DATE'].astype(int).astype(str)
    industry_info['END_DATE'] = industry_info['END_DATE'].astype(int).astype(str)
    industry_info['INDUSTRY_VERSION'] = industry_info['INDUSTRY_VERSION'].astype(int)
    industry_info['INDUSTRY_TYPE'] = industry_info['INDUSTRY_TYPE'].astype(int)
    industry_info['IS_NEW'] = industry_info['IS_NEW'].astype(int)
    industry_info = industry_info[industry_info['INDUSTRY_VERSION'] == 3]
    return industry_info

def get_industry_symbol():
    industry_symbol = HBDB().read_industry_symbol()
    industry_symbol = industry_symbol.rename(columns={'hyhfbz': 'INDUSTRY_VERSION', 'fldm': 'INDUSTRY_ID', 'flmc': 'INDUSTRY_NAME', 'zsdm': 'INDEX_SYMBOL', 'fljb': 'INDUSTRY_TYPE', 'qsrq': 'BEGIN_DATE', 'jsrq': 'END_DATE', 'sfyx': 'IS_NEW'})
    industry_symbol = industry_symbol.dropna(subset=['BEGIN_DATE'])
    industry_symbol['END_DATE'] = industry_symbol['END_DATE'].replace('', np.nan).fillna('20990101')
    industry_symbol['BEGIN_DATE'] = industry_symbol['BEGIN_DATE'].astype(int).astype(str)
    industry_symbol['END_DATE'] = industry_symbol['END_DATE'].astype(int).astype(str)
    industry_symbol['INDUSTRY_VERSION'] = industry_symbol['INDUSTRY_VERSION'].astype(int)
    industry_symbol['INDUSTRY_TYPE'] = industry_symbol['INDUSTRY_TYPE'].astype(int)
    industry_symbol['IS_NEW'] = industry_symbol['IS_NEW'].astype(int)
    industry_symbol = industry_symbol[industry_symbol['INDUSTRY_VERSION'] == 3]
    return industry_symbol

def get_stock_industry():
    stock_industry = HBDB().read_stock_industry()
    stock_industry.to_hdf('{0}stock_industry.hdf'.format(data_path), key='table', mode='w')
    stock_industry = pd.read_hdf('{0}stock_industry.hdf'.format(data_path), key='table')
    stock_industry = stock_industry.rename(columns={'zqdm': 'TICKER_SYMBOL', 'flmc': 'INDUSTRY_NAME', 'fldm': 'INDUSTRY_ID', 'hyhfbz': 'INDUSTRY_VERSION', 'fljb': 'INDUSTRY_TYPE', 'qsrq': 'BEGIN_DATE', 'jsrq': 'END_DATE', 'sfyx': 'IS_NEW'})
    stock_industry = stock_industry.dropna(subset=['BEGIN_DATE'])
    stock_industry['END_DATE'] = stock_industry['END_DATE'].fillna('20990101')
    stock_industry['BEGIN_DATE'] = stock_industry['BEGIN_DATE'].astype(int).astype(str)
    stock_industry['END_DATE'] = stock_industry['END_DATE'].astype(int).astype(str)
    stock_industry['INDUSTRY_VERSION'] = stock_industry['INDUSTRY_VERSION'].astype(int)
    stock_industry['INDUSTRY_TYPE'] = stock_industry['INDUSTRY_TYPE'].astype(int)
    stock_industry['IS_NEW'] = stock_industry['IS_NEW'].astype(int)
    stock_industry = stock_industry[stock_industry['INDUSTRY_VERSION'] == 2]
    stock_industry = stock_industry.loc[stock_industry['TICKER_SYMBOL'].str.len() == 6]
    stock_industry = stock_industry.loc[stock_industry['TICKER_SYMBOL'].astype(str).str.slice(0, 1).isin(['0', '3', '6'])]
    return stock_industry

def filter_extreme_mad(ser, n=3):
    median = ser.quantile(0.5)
    new_median = ((ser - median).abs()).quantile(0.5)
    max_range = median + n * new_median
    min_range = median - n * new_median
    ser = np.clip(ser, min_range, max_range)
    return ser


# 大中小盘统计分析
data_path = 'D:/Git/hbshare/hbshare/fe/xwq/data/jc_comparasion/'
calendar_df, report_df, trade_df, report_trade_df, calendar_trade_df = get_date('20150101', '20221231')
dates = report_trade_df['TRADE_DATE'].unique().tolist()[3:-2]
st = get_st('20150101', '20221231')
stock_info = get_stock_info()
stock_daily_k = pd.read_hdf('D:/Git/hbshare/hbshare/fe/xwq/data/jc_comparasion/stock_daily_k.hdf', key='table')
stock_daily_k = stock_daily_k[['TDATE', 'SYMBOL', 'TCAP']]
stock_daily_k.columns = ['TRADE_DATE', 'TICKER_SYMBOL', 'TCAP']
stock_daily_k['TRADE_DATE'] = stock_daily_k['TRADE_DATE'].astype(str)
stock_daily_k['TICKER_SYMBOL'] = stock_daily_k['TICKER_SYMBOL'].astype(str)
stock_daily_k = stock_daily_k.loc[stock_daily_k['TICKER_SYMBOL'].str.len() == 6]
stock_daily_k = stock_daily_k.loc[stock_daily_k['TICKER_SYMBOL'].astype(str).str.slice(0, 1).isin(['0', '3', '6'])]
stock_daily_k_mean = stock_daily_k.pivot(index='TRADE_DATE', columns='TICKER_SYMBOL', values='TCAP')
stock_daily_k_mean = stock_daily_k_mean.sort_index()
stock_daily_k_mean = stock_daily_k_mean.rolling(120).mean()
stock_daily_k_mean = stock_daily_k_mean.unstack().reset_index()
stock_daily_k_mean.columns = ['TICKER_SYMBOL', 'TRADE_DATE', 'TCAP_MEAN']

size_index = FEDB().read_size_index()
size_index = size_index[size_index['asofdate'] >= '201503']
size_index = size_index.merge(report_trade_df[['YEAR_MONTH', 'TRADE_DATE']].rename(columns={'YEAR_MONTH': 'asofdate'}), on=['asofdate'], how='left')
size_index = size_index[['TRADE_DATE', '代码', 'type']]
size_index.columns = ['TRADE_DATE', 'TICKER_SYMBOL', 'MARK']

jc_thred_list, market_thred_list = [], []
for date in dates:
    print(date)
    stock_daily_k_mean_date = stock_daily_k_mean[stock_daily_k_mean['TRADE_DATE'] == date]
    st_date = st.loc[date].dropna()
    stock_daily_k_mean_date = stock_daily_k_mean_date[~stock_daily_k_mean_date['TICKER_SYMBOL'].isin(list(st_date.index))]
    stock_info_date = stock_info[stock_info['ESTABLISH_DATE'] < (datetime.strptime(date, '%Y%m%d') - timedelta(180)).strftime('%Y%m%d')]
    stock_daily_k_mean_date = stock_daily_k_mean_date[stock_daily_k_mean_date['TICKER_SYMBOL'].isin(stock_info_date['TICKER_SYMBOL'].unique().tolist())]
    # 巨潮阈值
    jc_thred = size_index[size_index['TRADE_DATE'] == date]
    jc_thred = jc_thred.merge(stock_daily_k_mean_date, on=['TRADE_DATE', 'TICKER_SYMBOL'], how='left')
    large_middle_thred = jc_thred['TCAP_MEAN'].quantile(0.8)
    middle_small_thred = jc_thred['TCAP_MEAN'].quantile(0.5)
    jc_thred = jc_thred.set_index('TICKER_SYMBOL').reindex(stock_daily_k_mean_date['TICKER_SYMBOL'].unique().tolist()).drop('TCAP_MEAN', axis=1)
    jc_thred['TRADE_DATE'] = jc_thred['TRADE_DATE'].fillna(date)
    jc_thred = jc_thred.merge(stock_daily_k_mean_date, on=['TRADE_DATE', 'TICKER_SYMBOL'], how='left')
    jc_thred_not_isnan = jc_thred[~jc_thred['MARK'].isnull()]
    jc_thred_isnan = jc_thred[jc_thred['MARK'].isnull()]
    jc_thred_isnan['MARK'] = '中盘'
    jc_thred_isnan.loc[jc_thred_isnan['TCAP_MEAN'] > large_middle_thred, 'MARK'] = '大盘'
    jc_thred_isnan.loc[jc_thred_isnan['TCAP_MEAN'] < middle_small_thred, 'MARK'] = '小盘'
    jc_thred = pd.concat([jc_thred_not_isnan, jc_thred_isnan])
    jc_thred_count = jc_thred[['MARK', 'TICKER_SYMBOL']].groupby('MARK').count().rename(columns={'TICKER_SYMBOL': 'COUNT'})
    jc_thred_min = jc_thred[['MARK', 'TCAP_MEAN']].groupby('MARK').min().rename(columns={'TCAP_MEAN': 'MIN'})
    jc_thred_mean = jc_thred[['MARK', 'TCAP_MEAN']].groupby('MARK').mean().rename(columns={'TCAP_MEAN': 'MEAN'})
    jc_thred_max = jc_thred[['MARK', 'TCAP_MEAN']].groupby('MARK').max().rename(columns={'TCAP_MEAN': 'MAX'})
    jc_thred = pd.concat([jc_thred_count, jc_thred_min, jc_thred_mean, jc_thred_max], axis=1)
    jc_thred = jc_thred.loc[['大盘', '中盘', '小盘']]
    jc_thred = jc_thred.reset_index()
    jc_thred['TRADE_DATE'] = date
    jc_thred = jc_thred.set_index(['TRADE_DATE', 'MARK'])
    jc_thred_list.append(jc_thred)
    # 全市场阈值
    market_thred = stock_daily_k_mean_date.sort_values('TCAP_MEAN', ascending=False).dropna()
    market_thred['MARK'] = '中盘'
    market_thred['MARK'].iloc[:int(len(market_thred) * 0.2)] = '大盘'
    market_thred['MARK'].iloc[-int(len(market_thred) * 0.5):] = '小盘'
    market_thred_count = market_thred[['MARK', 'TICKER_SYMBOL']].groupby('MARK').count().rename(columns={'TICKER_SYMBOL': 'COUNT'})
    market_thred_min = market_thred[['MARK', 'TCAP_MEAN']].groupby('MARK').min().rename(columns={'TCAP_MEAN': 'MIN'})
    market_thred_mean = market_thred[['MARK', 'TCAP_MEAN']].groupby('MARK').mean().rename(columns={'TCAP_MEAN': 'MEAN'})
    market_thred_max = market_thred[['MARK', 'TCAP_MEAN']].groupby('MARK').max().rename(columns={'TCAP_MEAN': 'MAX'})
    market_thred = pd.concat([market_thred_count, market_thred_min, market_thred_mean, market_thred_max], axis=1)
    market_thred = market_thred.loc[['大盘', '中盘', '小盘']]
    market_thred = market_thred.reset_index()
    market_thred['TRADE_DATE'] = date
    market_thred = market_thred.set_index(['TRADE_DATE', 'MARK'])
    market_thred_list.append(market_thred)
all_jc_thred = pd.concat(jc_thred_list)
all_market_thred = pd.concat(market_thred_list)
all_jc_thred.to_excel('{0}all_jc_thred.xlsx'.format(data_path))
all_market_thred.to_excel('{0}all_market_thred.xlsx'.format(data_path))

# data = all_jc_thred.copy(deep=True)
# data = data.reset_index()
# data = data.pivot(index='TRADE_DATE', columns='MARK', values='MEAN')
# data.to_excel('{0}all_jc_thred_mean.xlsx'.format(data_path))


# 成长价值统计分析
data_path = 'D:/Git/hbshare/hbshare/fe/xwq/data/jc_comparasion/'
calendar_df, report_df, trade_df, report_trade_df, calendar_trade_df = get_date('20150101', '20221231')
dates = report_trade_df['TRADE_DATE'].unique().tolist()[3:-2]
st = get_st('20150101', '20221231')
stock_info = get_stock_info()

stock_daily_k = pd.read_hdf('D:/Git/hbshare/hbshare/fe/xwq/data/jc_comparasion/stock_daily_k.hdf', key='table')
stock_daily_k = stock_daily_k[['TDATE', 'SYMBOL', 'TCLOSE', 'TCAP']]
stock_daily_k.columns = ['TRADE_DATE', 'TICKER_SYMBOL', 'CLOSE_PRICE', 'MARKET_VALUE']
stock_daily_k['TRADE_DATE'] = stock_daily_k['TRADE_DATE'].astype(str)
stock_daily_k['TICKER_SYMBOL'] = stock_daily_k['TICKER_SYMBOL'].astype(str)
stock_daily_k = stock_daily_k.loc[stock_daily_k['TICKER_SYMBOL'].str.len() == 6]
stock_daily_k = stock_daily_k.loc[stock_daily_k['TICKER_SYMBOL'].astype(str).str.slice(0, 1).isin(['0', '3', '6'])]

stock_finance = pd.read_hdf('D:/Git/hbshare/hbshare/fe/xwq/data/jc_comparasion/stock_finance.hdf', key='table')
stock_finance = stock_finance.sort_values(['TICKER_SYMBOL', 'END_DATE', 'PUBLISH_DATE']).drop_duplicates(['TICKER_SYMBOL', 'END_DATE'], keep='last')
stock_finance = stock_finance.merge(calendar_trade_df[['CALENDAR_DATE', 'TRADE_DATE']].rename(columns={'CALENDAR_DATE': 'END_DATE'}), on=['END_DATE'], how='left')
stock_finance = stock_finance.merge(stock_daily_k[['TICKER_SYMBOL', 'TRADE_DATE', 'CLOSE_PRICE', 'MARKET_VALUE']], on=['TICKER_SYMBOL', 'TRADE_DATE'], how='left')
stock_finance['CLOSE_PRICE'] = stock_finance['CLOSE_PRICE'].replace(0.0, np.nan)
stock_finance['OPER_REVENUE_YOY_MEAN'] = stock_finance['OPER_REVENUE_YOY'].rolling(12).mean()
stock_finance['NET_PROFIT_YOY_MEAN'] = stock_finance['NET_PROFIT_YOY'].rolling(12).mean()
stock_finance['EP_TTM'] = stock_finance['EPS_TTM'] / stock_finance['CLOSE_PRICE']
stock_finance['BP'] = stock_finance['NAPS'] / stock_finance['CLOSE_PRICE']
stock_finance['CFP_TTM'] = stock_finance['OCF_TTM'] / stock_finance['CLOSE_PRICE']
stock_finance['DP_TTM'] = stock_finance['DIVIDEND_TTM'] / stock_finance['MARKET_VALUE']
stock_finance = stock_finance[['TICKER_SYMBOL', 'END_DATE', 'OPER_REVENUE_YOY_MEAN', 'NET_PROFIT_YOY_MEAN', 'ROE_TTM', 'EP_TTM', 'BP', 'CFP_TTM', 'DP_TTM']]
industry_info = get_industry_info()
industry_info = industry_info[industry_info['INDUSTRY_TYPE'] == 1]
industry_info = industry_info[industry_info['IS_NEW'] == 1]
industry_info = industry_info[['INDUSTRY_ID', 'INDUSTRY_NAME']]
industry_symbol = get_industry_symbol()
industry_symbol = industry_symbol[industry_symbol['INDUSTRY_TYPE'] == 1]
industry_symbol = industry_symbol[industry_symbol['IS_NEW'] == 1]
industry_symbol = industry_symbol[['INDUSTRY_ID', 'INDUSTRY_NAME']]
stock_industry = get_stock_industry()
stock_industry = stock_industry[stock_industry['INDUSTRY_TYPE'] == 1]
stock_industry = stock_industry[stock_industry['IS_NEW'] == 1]
stock_industry = stock_industry.drop('INDUSTRY_NAME', axis=1).merge(industry_symbol, on=['INDUSTRY_ID'], how='left')
stock_industry = stock_industry.drop('INDUSTRY_ID', axis=1).merge(industry_info[['INDUSTRY_NAME', 'INDUSTRY_ID']], on=['INDUSTRY_NAME'], how='left')
stock_industry = stock_industry[['INDUSTRY_ID', 'INDUSTRY_NAME', 'TICKER_SYMBOL', 'BEGIN_DATE', 'END_DATE']]
stock_finance = stock_finance.merge(stock_industry[['TICKER_SYMBOL', 'INDUSTRY_ID']], on=['TICKER_SYMBOL'], how='left')
stock_finance_list = []
for date in stock_finance['END_DATE'].unique().tolist():
    stock_finance_date = stock_finance[stock_finance['END_DATE'] == date]
    for col in ['OPER_REVENUE_YOY_MEAN', 'NET_PROFIT_YOY_MEAN', 'ROE_TTM', 'EP_TTM', 'BP', 'CFP_TTM', 'DP_TTM']:
        stock_finance_date[col] = filter_extreme_mad(stock_finance_date[col])
        industry_finance = stock_finance_date[['INDUSTRY_ID', col]].groupby('INDUSTRY_ID').mean().rename(columns={col: col + '_INDU'})
        stock_finance_date = stock_finance_date.merge(industry_finance, on=['INDUSTRY_ID'], how='left')
        stock_finance_date[col] = stock_finance_date.apply(lambda x: x[col + '_INDU'] if np.isnan(x[col]) else x[col], axis=1)
        stock_finance_date[col] = (stock_finance_date[col] - stock_finance_date[col].mean()) / stock_finance_date[col].std(ddof=1)
    stock_finance_list.append(stock_finance_date)
    print(date)
stock_finance = pd.concat(stock_finance_list)
stock_finance = stock_finance[['TICKER_SYMBOL', 'END_DATE', 'OPER_REVENUE_YOY_MEAN', 'NET_PROFIT_YOY_MEAN', 'ROE_TTM', 'EP_TTM', 'BP', 'CFP_TTM', 'DP_TTM']]
stock_finance['GROWTH'] = (stock_finance['OPER_REVENUE_YOY_MEAN'] + stock_finance['NET_PROFIT_YOY_MEAN'] + stock_finance['ROE_TTM']) / 3.0
stock_finance['VALUE'] = (stock_finance['EP_TTM'] + stock_finance['BP'] + stock_finance['CFP_TTM'] + stock_finance['DP_TTM']) / 4.0
# 成长因子
growth = stock_finance.pivot(index='END_DATE', columns='TICKER_SYMBOL', values='GROWTH').sort_index()
growth = calendar_df[['CALENDAR_DATE']].set_index('CALENDAR_DATE').merge(growth, left_index=True, right_index=True, how='left')
growth = growth.fillna(method='ffill').dropna(how='all')
growth = growth[growth.index.isin(trade_df['TRADE_DATE'].unique())]
growth = growth.unstack().reset_index()
growth.columns = ['TICKER_SYMBOL', 'TRADE_DATE', 'GROWTH']
# 价值因子
value = stock_finance.pivot(index='END_DATE', columns='TICKER_SYMBOL', values='VALUE').sort_index()
value = calendar_df[['CALENDAR_DATE']].set_index('CALENDAR_DATE').merge(value, left_index=True, right_index=True, how='left')
value = value.fillna(method='ffill').dropna(how='all')
value = value[value.index.isin(trade_df['TRADE_DATE'].unique())]
value = value.unstack().reset_index()
value.columns = ['TICKER_SYMBOL', 'TRADE_DATE', 'VALUE']
growth.to_hdf('{0}growth.hdf'.format(data_path), key='table', mode='w')
value.to_hdf('{0}value.hdf'.format(data_path), key='table', mode='w')
growth = pd.read_hdf('{0}growth.hdf'.format(data_path), key='table')
value = pd.read_hdf('{0}value.hdf'.format(data_path), key='table')

style_index = FEDB().read_style_index()
style_index = style_index[style_index['asofdate'] >= '201503']
style_index = style_index.merge(report_trade_df[['YEAR_MONTH', 'TRADE_DATE']].rename(columns={'YEAR_MONTH': 'asofdate'}), on=['asofdate'], how='left')
style_index = style_index[['TRADE_DATE', '代码', 'type']]
style_index.columns = ['TRADE_DATE', 'TICKER_SYMBOL', 'MARK']

jc_thred_list, market_thred_list = [], []
for date in dates:
    print(date)
    stock_daily_k_date = stock_daily_k[stock_daily_k['TRADE_DATE'] == date]
    st_date = st.loc[date].dropna()
    stock_daily_k_date = stock_daily_k_date[~stock_daily_k_date['TICKER_SYMBOL'].isin(list(st_date.index))]
    stock_info_date = stock_info[stock_info['ESTABLISH_DATE'] < (datetime.strptime(date, '%Y%m%d') - timedelta(180)).strftime('%Y%m%d')]
    stock_daily_k_date = stock_daily_k_date[stock_daily_k_date['TICKER_SYMBOL'].isin(stock_info_date['TICKER_SYMBOL'].unique().tolist())]
    # 巨潮阈值
    jc_thred = style_index[style_index['TRADE_DATE'] == date]
    jc_thred = jc_thred.merge(growth, on=['TRADE_DATE', 'TICKER_SYMBOL'], how='left').merge(value, on=['TRADE_DATE', 'TICKER_SYMBOL'], how='left')
    growth_thred = jc_thred['GROWTH'].quantile(0.67)
    value_thred = jc_thred['VALUE'].quantile(0.67)
    jc_thred = jc_thred.drop_duplicates('TICKER_SYMBOL').set_index('TICKER_SYMBOL').reindex(stock_daily_k_date['TICKER_SYMBOL'].unique().tolist()).drop(['GROWTH', 'VALUE'], axis=1)
    jc_thred['TRADE_DATE'] = jc_thred['TRADE_DATE'].fillna(date)
    jc_thred = jc_thred.merge(stock_daily_k_date, on=['TRADE_DATE', 'TICKER_SYMBOL'], how='left').merge(growth, on=['TRADE_DATE', 'TICKER_SYMBOL'], how='left').merge(value, on=['TRADE_DATE', 'TICKER_SYMBOL'], how='left')
    jc_thred_not_isnan = jc_thred[~jc_thred['MARK'].isnull()]
    jc_thred_isnan = jc_thred[jc_thred['MARK'].isnull()]
    jc_thred_isnan['MARK'] = np.nan
    jc_thred_isnan.loc[(jc_thred_isnan['GROWTH'] > growth_thred) & (jc_thred_isnan['VALUE'] <= value_thred), 'MARK'] = '成长'
    jc_thred_isnan.loc[(jc_thred_isnan['VALUE'] > value_thred) & (jc_thred_isnan['GROWTH'] <= growth_thred), 'MARK'] = '价值'
    jc_thred = pd.concat([jc_thred_not_isnan, jc_thred_isnan])
    jc_thred = jc_thred.dropna(subset=['MARK'])
    jc_thred_count = jc_thred[['MARK', 'TICKER_SYMBOL']].groupby('MARK').count().rename(columns={'TICKER_SYMBOL': 'COUNT'})
    jc_thred_min = jc_thred[['MARK', 'MARKET_VALUE']].groupby('MARK').min().rename(columns={'MARKET_VALUE': 'MIN'})
    jc_thred_mean = jc_thred[['MARK', 'MARKET_VALUE']].groupby('MARK').mean().rename(columns={'MARKET_VALUE': 'MEAN'})
    jc_thred_max = jc_thred[['MARK', 'MARKET_VALUE']].groupby('MARK').max().rename(columns={'MARKET_VALUE': 'MAX'})
    jc_thred = pd.concat([jc_thred_count, jc_thred_min, jc_thred_mean, jc_thred_max], axis=1)
    jc_thred = jc_thred.loc[['成长', '价值']]
    jc_thred = jc_thred.reset_index()
    jc_thred['TRADE_DATE'] = date
    jc_thred = jc_thred.set_index(['TRADE_DATE', 'MARK'])
    jc_thred_list.append(jc_thred)
    # 全市场阈值
    market_thred = stock_daily_k_date.merge(growth, on=['TRADE_DATE', 'TICKER_SYMBOL'], how='left').merge(value, on=['TRADE_DATE', 'TICKER_SYMBOL'], how='left')
    growth_thred = market_thred['GROWTH'].quantile(0.67)
    value_thred = market_thred['VALUE'].quantile(0.67)
    market_thred['MARK'] = np.nan
    market_thred.loc[(market_thred['GROWTH'] > growth_thred) & (market_thred['VALUE'] <= value_thred), 'MARK'] = '成长'
    market_thred.loc[(market_thred['VALUE'] > value_thred) & (market_thred['GROWTH'] <= growth_thred), 'MARK'] = '价值'
    market_thred = market_thred.dropna(subset=['MARK'])
    market_thred_count = market_thred[['MARK', 'TICKER_SYMBOL']].groupby('MARK').count().rename(columns={'TICKER_SYMBOL': 'COUNT'})
    market_thred_min = market_thred[['MARK', 'MARKET_VALUE']].groupby('MARK').min().rename(columns={'MARKET_VALUE': 'MIN'})
    market_thred_mean = market_thred[['MARK', 'MARKET_VALUE']].groupby('MARK').mean().rename(columns={'MARKET_VALUE': 'MEAN'})
    market_thred_max = market_thred[['MARK', 'MARKET_VALUE']].groupby('MARK').max().rename(columns={'MARKET_VALUE': 'MAX'})
    market_thred = pd.concat([market_thred_count, market_thred_min, market_thred_mean, market_thred_max], axis=1)
    market_thred = market_thred.loc[['成长', '价值']]
    market_thred = market_thred.reset_index()
    market_thred['TRADE_DATE'] = date
    market_thred = market_thred.set_index(['TRADE_DATE', 'MARK'])
    market_thred_list.append(market_thred)
all_jc_thred = pd.concat(jc_thred_list)
all_market_thred = pd.concat(market_thred_list)
all_jc_thred.to_excel('{0}all_jc_thred.xlsx'.format(data_path))
all_market_thred.to_excel('{0}all_market_thred.xlsx'.format(data_path))

# data = all_jc_thred.copy(deep=True)
# data = data.reset_index()
# data = data.pivot(index='TRADE_DATE', columns='MARK', values='MEAN')
# data.to_excel('{0}all_jc_thred_mean.xlsx'.format(data_path))





"""
市场中性策略模块
"""
import pandas as pd
import numpy as np
import hbshare as hbs
from datetime import datetime
from hbshare.quant.Kevin.quant_room.MyUtil.data_loader import get_trading_day_list, get_fund_nav_from_sql
from Arbitrage_backtest import cal_annual_return, cal_annual_volatility, cal_sharpe_ratio, cal_max_drawdown


class NeutralTracker:
    def __init__(self, start_date, end_date, config_df):
        self.start_date = start_date
        self.end_date = end_date
        self.config_df = config_df
        self._load_data()

    def _load_data(self):
        config_df = self.config_df.copy()
        config_df['运作日期'] = config_df['运作日期'].apply(lambda x: datetime.strftime(x, '%Y%m%d'))
        config_df['指增起始'] = config_df['指增起始'].apply(lambda x: datetime.strftime(x, '%Y%m%d'))
        neu_fund_dict = self.config_df.set_index('管理人')['基金代码'].to_dict()
        # 中性
        neu_start = config_df['运作日期'].min()
        neu_date_list = get_trading_day_list(neu_start, self.end_date, frequency="week")
        fund_nav = get_fund_nav_from_sql(neu_start, self.end_date, neu_fund_dict).reindex(neu_date_list).dropna(how='all')
        for fund_name in fund_nav.columns:
            act_start = config_df.set_index('管理人').loc[fund_name]['运作日期']
            fund_nav.loc[:act_start, fund_name] = np.NaN
        neu_nav = fund_nav.dropna(how='all').fillna(method='ffill')
        # 指增
        alpha_fund_dict = self.config_df.set_index('管理人')['对应指增'].to_dict()
        alpha_start = config_df['指增起始'].min()
        alpha_date_list = get_trading_day_list(alpha_start, self.end_date, frequency="week")
        fund_nav = get_fund_nav_from_sql(alpha_start, self.end_date, alpha_fund_dict).reindex(alpha_date_list).dropna(
            how='all')
        for fund_name in fund_nav.columns:
            act_start = config_df.set_index('管理人').loc[fund_name]['指增起始']
            fund_nav.loc[:act_start, fund_name] = np.NaN
        alpha_nav = fund_nav.dropna(how='all').fillna(method='ffill')

        self.neu_nav = neu_nav
        self.alpha_nav = alpha_nav

    def calculate_neu_ratio(self):
        nav_df = self.neu_nav.dropna()

        portfolio_index_df = pd.DataFrame(
            index=nav_df.columns, columns=['超额年化收益', '超额年化波动', '最大回撤', 'Sharpe', '胜率', '平均损益比',
                                           'period_1', 'period_2', 'period_3'])
        portfolio_index_df.loc[:, '超额年化收益'] = nav_df.pct_change().dropna(how='all').apply(cal_annual_return, axis=0)
        portfolio_index_df.loc[:, '超额年化波动'] = \
            nav_df.pct_change().dropna(how='all').apply(cal_annual_volatility, axis=0)
        portfolio_index_df.loc[:, '最大回撤'] = nav_df.apply(cal_max_drawdown, axis=0)
        portfolio_index_df.loc[:, 'Sharpe'] = \
            nav_df.pct_change().dropna(how='all').apply(lambda x: cal_sharpe_ratio(x, 0.015), axis=0)
        portfolio_index_df.loc[:, '胜率'] = \
            nav_df.pct_change().dropna(how='all').apply(lambda x: x.gt(0).sum() / len(x), axis=0)
        portfolio_index_df.loc[:, '平均损益比'] = \
            nav_df.pct_change().dropna(how='all').apply(lambda x: x[x > 0].mean() / x[x < 0].abs().mean(), axis=0)
        portfolio_index_df.loc[:, "period_1"] = nav_df.loc["20210917"] / nav_df.loc["20210416"] - 1
        portfolio_index_df.loc[:, "period_2"] = nav_df.loc["20220211"] / nav_df.loc["20210917"] - 1
        portfolio_index_df.loc[:, "period_3"] = nav_df.loc["20220916"] / nav_df.loc["20220211"] - 1

        portfolio_index_df.index.name = '产品名称'
        portfolio_index_df.reset_index(inplace=True)
        # 格式处理
        portfolio_index_df['超额年化收益'] = portfolio_index_df['超额年化收益'].apply(lambda x: format(x, '.2%'))
        portfolio_index_df['超额年化波动'] = portfolio_index_df['超额年化波动'].apply(lambda x: format(x, '.2%'))
        portfolio_index_df['最大回撤'] = portfolio_index_df['最大回撤'].apply(lambda x: format(x, '.2%'))
        portfolio_index_df['Sharpe'] = portfolio_index_df['Sharpe'].round(2)
        portfolio_index_df['胜率'] = portfolio_index_df['胜率'].apply(lambda x: format(x, '.1%'))
        portfolio_index_df['平均损益比'] = portfolio_index_df['平均损益比'].round(2)
        portfolio_index_df['period_1'] = portfolio_index_df['period_1'].apply(lambda x: format(x, '.2%'))
        portfolio_index_df['period_2'] = portfolio_index_df['period_2'].apply(lambda x: format(x, '.2%'))
        portfolio_index_df['period_3'] = portfolio_index_df['period_3'].apply(lambda x: format(x, '.2%'))

        return portfolio_index_df

    def hedge_compare(self, start_date="20210806"):
        nav_df = self.neu_nav.dropna().loc[start_date:]
        enhance_df = self.alpha_nav.loc[start_date:]

        sql_script = "SELECT JYRQ as TRADEDATE, ZQDM, SPJG as TCLOSE from funddb.ZSJY WHERE" \
                     " ZQDM = '{}' " \
                     "and JYRQ >= {} and JYRQ <= {}".format("000905", start_date, enhance_df.index[-1])
        res = hbs.db_data_query('readonly', sql_script, page_size=5000)
        data = pd.DataFrame(res['data'])
        benchmark_series = data.set_index('TRADEDATE')['TCLOSE'].reindex(enhance_df.index)

        alpha_df = enhance_df.pct_change().dropna().sub(benchmark_series.pct_change().dropna(), axis=0).sort_index()
        neutral_df = nav_df.pct_change().dropna()

        assert (alpha_df.shape == neutral_df.shape)

        hedge_df = 1.2 * neutral_df - alpha_df

        # 收益拆解
        attr_list = []
        for col in alpha_df.columns:
            kt = np.log(1 + neutral_df[col]) / neutral_df[col]
            r = (1 + neutral_df[col]).prod() - 1
            k = np.log(1 + r) / r

            attr = alpha_df[col].to_frame('超额').merge(
                hedge_df[col].to_frame('对冲'), left_index=True, right_index=True) / 1.2

            cum_attr = attr.T.multiply(kt / k).sum(axis=1)

            attr_list.append(cum_attr.to_frame(col))

        attr_df = pd.concat(attr_list, axis=1).T

        # 风险拆解
        risk_list = []
        for col in alpha_df.columns:
            period_df = alpha_df[col].to_frame('超额').merge(
                hedge_df[col].to_frame('对冲'), left_index=True, right_index=True)
            cov_df = period_df.cov() * 10000.
            V = np.matrix(cov_df)
            w = np.matrix(np.array([1, 1]))
            sigma = np.sqrt((w * V * w.T)[0, 0])
            MRC = V * w.T
            # Risk Contribution
            RC = np.multiply(MRC, w.T) / sigma / 120

            attr = pd.DataFrame(index=['超额', '对冲'], columns=[col], data=RC.flatten().tolist()[0])
            risk_list.append(attr)

        risk_attr = pd.concat(risk_list, axis=1) * np.sqrt(52)
        risk_attr = risk_attr.T

        return attr_df, risk_attr


if __name__ == '__main__':
    config = pd.read_excel("D:\\研究基地\\G-专题报告\\【2022.9】市场中性专题（小专题）\\专题报告数据.xlsx", sheet_name=0)

    NeutralTracker('20200101', '20221230', config).hedge_compare()


import numpy as np
from hbshare.quant.CChen.fund_stats.perf import performance_analysis, performance_specific_ret
from hbshare.quant.CChen.data.data import load_funds_data, load_funds_alpha
import pandas as pd
from hbshare.quant.CChen.db_const import sql_write_path_work
from datetime import timedelta
import matplotlib.pyplot as plt
from pandas.plotting import register_matplotlib_converters
import riskfolio as rp

register_matplotlib_converters()
plt.rcParams['font.family'] = ['sans-serif']
plt.rcParams['font.sans-serif'] = ['KaiTi']

pd.set_option('display.unicode.ambiguous_as_wide', True)
pd.set_option('display.unicode.east_asian_width', True)
pd.set_option('display.width', 5500)


class FOFBacktest(object):

    investable_date = None
    db_path = sql_write_path_work['work']
    # cal_path = sql_write_path_work['daily']

    result = None
    fof_result = None

    trade_dict = {
        1: [
            1,
            # 4,
            7,
            # 10
        ],
        2: [2, 5, 8, 11],
        3: [3, 6, 9, 12],
    }

    def __init__(
            self,
            start_date,
            end_date,
            targets,
            date_key='t_date',
            alpha=False,
            data=None,
            trade_m=1
    ):
        self.start_date = start_date
        self.end_date = end_date
        self.targets = targets
        self.target_weight = None
        self.tm = trade_m

        self.funds = pd.read_sql_query(
            'select * from fund_list where `name` in' + str(tuple(self.targets['name'].tolist())).replace(',)', ')'),
            self.db_path
        )

        if data is None:
            self.data = load_funds_data(
                fund_list=self.funds,
                first_date=start_date - timedelta(days=365 * 2),
                end_date=end_date,
                freq='w',
                db_path=self.db_path,
                # cal_db_path=self.cal_path,
                drop_1=True,
            )[[date_key] + self.targets['name'].tolist()]
        else:
            self.data = data

        if alpha:
            self.data_alpha = load_funds_alpha(
                fund_list=self.funds,
                first_date=start_date - timedelta(days=365 * 2),
                end_date=end_date,
                freq='w',
                db_path=self.db_path,
                # cal_db_path=self.cal_path
            )['eav']

        self.calendar = self.data[date_key].tolist()
        self.data = self.data.set_index(self.data[date_key]).ffill()
        self.data = self.data[self.targets['name'].tolist()]
        self.data_return = self.data.pct_change()

        self.fund_pos_df = pd.DataFrame()
        self.fund_value_df = pd.DataFrame()
        self.fund_contribution_df = pd.DataFrame()
        self.equity = pd.DataFrame()

        self.money = 1

    def current_targets(self, date):
        fund_pool = self.targets['name'].tolist().copy()
        fund_targets = self.data.loc[date, fund_pool]
        return fund_targets

    def consider_investable_date(self, date, fund_pool):
        # 考虑入池时间
        if self.investable_date is not None:
            for f in self.investable_date:
                if date < self.investable_date[f]['入池时间'].date():
                    if f in fund_pool:
                        fund_pool.remove(f)
        return fund_pool

    def compute_initial_weights(self, date, fund_targets):
        pass

    def compute_weights(self, date, fund_targets):
        pass

    def compute_contribution(self, date):
        date_lag = self.calendar[self.calendar.index(date) - 1]
        funds_return = self.current_targets(date=date) - self.current_targets(date=date_lag)
        fund_contribution = funds_return * self.fund_pos_df.loc[date_lag]
        self.fund_contribution_df = pd.concat(
            [
                self.fund_contribution_df,
                pd.DataFrame(fund_contribution).T.rename({0: date})
            ],
            sort=True
        )

    def trade_status(self, current_date):
        # 判定是否交易

        if_trade = (
                current_date.month in self.trade_dict[self.tm]
                and current_date.month != self.calendar[self.calendar.index(current_date) - 1].month
        )
        return if_trade

    def run(self, no_adjust_untill_new=False, trade_lag=False):
        start_index = sum(np.array(self.calendar) < self.start_date)
        fund_pos = None
        self.equity = pd.DataFrame(
            {
                '现金': 0
            }, index=[self.calendar[start_index - 1]]
        )

        for i in range(start_index, len(self.calendar)):
            current_date = self.calendar[i]
            # print(current_date)
            fund_targets = self.current_targets(date=current_date)

            if i == start_index:
                # 起点建仓
                fund_pos = self.compute_initial_weights(date=current_date, fund_targets=fund_targets)
                fund_value = fund_pos.mul(fund_targets)
                self.fund_contribution_df = fund_value.copy()
                self.fund_contribution_df.iloc[:, :] = np.nan

            else:
                fund_pos = fund_pos.rename({self.data.index[i - 1]: self.data.index[i]})
                fund_value = fund_pos * fund_targets

                # 计算当期分产品贡献
                self.compute_contribution(date=current_date)

            self.fund_value_df = pd.concat([self.fund_value_df, fund_value], sort=True)
            if self.equity.loc[self.calendar[i - 1], '现金'] > 0:
                # 上一期总资产
                fund_pos = pos_future
                self.equity.loc[current_date, '现金'] = 0
            else:
                self.equity.loc[current_date, '现金'] = self.equity.loc[self.calendar[i - 1], '现金']

            # 交易
            if i > start_index and i != len(self.calendar) - 1:
                if self.trade_status(current_date=current_date):
                    if no_adjust_untill_new:
                        if int((~np.isnan(fund_value.values)).sum()) != len(fund_targets[fund_targets > 0]):
                            fund_pos = self.compute_weights(date=current_date, fund_targets=fund_targets)
                    else:
                        fund_pos = self.compute_weights(date=current_date, fund_targets=fund_targets)

                    if trade_lag:
                        # 考虑申购赎回间隔
                        position_before_trade = self.fund_pos_df.loc[
                            [self.calendar[i - 1]],
                            (self.fund_pos_df.loc[[self.calendar[i - 1]]] > 0).loc[self.calendar[i - 1]]
                            + (fund_pos.loc[[self.calendar[i]]] > 0).loc[self.calendar[i]]
                        ].fillna(0)  # 赎回前标的仓位
                        position_after_trade = fund_pos.loc[
                            [self.calendar[i]],
                            (self.fund_pos_df.loc[[self.calendar[i - 1]]] > 0).loc[self.calendar[i - 1]]
                            + (fund_pos.loc[[self.calendar[i]]] > 0).loc[self.calendar[i]]
                        ].fillna(0)  # 赎回后标的仓位
                        position_trade = position_before_trade.append(position_after_trade)
                        fund_pos = pd.DataFrame(position_trade.min(axis=0)).rename(columns={0: current_date}).loc[
                            (position_trade.loc[[self.calendar[i]]] > 0).loc[self.calendar[i]]
                        ].T  # 实际赎回操作后的仓位
                        fund_v = fund_pos * fund_targets
                        fund_v = fund_v[fund_v > 0]
                        # 计算目标持仓分产品持仓市值
                        # fund_value_trade = fund_pos.copy()
                        # fund_value_trade.loc[:, :] = self.fund_value_df.loc[current_date].sum() / len(fund_pos.columns)
                        fund_value_trade = position_after_trade * fund_targets
                        value_diff = fund_value_trade - fund_v[fund_value_trade.columns.tolist()].fillna(0)
                        # 下期持仓，涉及未来净值
                        pos_future = (
                                value_diff / self.current_targets(date=self.calendar[i + 1])[
                                    fund_value_trade.columns.tolist()
                                ] + fund_pos
                        ).set_index([[self.calendar[i + 1]]])
                        self.equity.loc[current_date, '现金'] = (
                                self.fund_value_df.loc[current_date].sum() - (fund_pos * fund_targets).sum(1)[0]
                        )

            self.fund_pos_df = pd.concat([self.fund_pos_df, fund_pos], sort=True)

            # if i == len(self.calendar) - 1:
            #     self.compute_contribution(date=current_date)

        if self.equity['现金'].max() > 0:
            self.fund_value_df['现金'] = self.equity['现金'][:-1].tolist()
        self.fof_result = pd.DataFrame(self.fund_value_df.sum(axis=1)).reset_index().rename(
            columns={'index': 't_date', 0: '组合'}
        )
        self.fof_result['动态回撤'] = (
                                        self.fof_result['组合'] - self.fof_result['组合'].cummax()
                                ) / self.fof_result['组合'].cummax()
        result1, _, _ = performance_analysis(
            data_df=self.fof_result[['t_date', '组合']],
            start_date=self.start_date,
            end_date=self.end_date,
            ret_num_per_year=52
        )
        result2 = performance_specific_ret(data_df=self.fof_result[['t_date', '组合']]).reset_index(drop=True)
        self.result = result1.append(result2.loc[6:, :])

    def summary(self, fig=True, title=''):
        pd.options.display.float_format = '{:.2f}'.format
        print(self.result[['index', '组合']].set_index('index'))
        print()
        pd.options.display.float_format = '{:.2%}'.format
        print(
            pd.DataFrame(
                self.fund_contribution_df.sum() / self.money
            ).sort_values(by=0, ascending=False).rename(columns={0: '贡献'})
        )
        # for i in self.fund_contribution_df.columns:
        #     print(i + ' 贡献：\n\t%.2f%%' % round(self.fund_contribution_df[i].sum() * 100 / self.money, 2))

        if fig:
            fig = plt.figure(figsize=(18, 8))
            ax1 = fig.add_subplot(1, 2, 1)
            ax1.plot(self.fof_result['t_date'], self.fof_result['组合'])
            plt.title(title)
            ax11 = ax1.twinx()
            ax11.fill_between(
                self.fof_result['t_date'],
                self.fof_result['动态回撤'],
                color='gray',
                alpha=0.3
            )

            ax2 = fig.add_subplot(1, 2, 2)
            plot_str = ''
            for i in self.fund_contribution_df.columns:
                plot_str += 'self.fund_value_df[\'' + i + '\'].fillna(0).tolist(), '
            eval(
                'ax2.stackplot(self.fund_value_df.index, '
                + plot_str
                + 'labels=list(self.fund_contribution_df.columns))'
            )

            # ax2.stackplot(
            #     fund_value_df.index, np.vstack(fund_value_df.fillna(0).values.T.tolist()),
            #     # labels=list(fund_name)
            # )

            plt.legend(loc=(0.95, 0.05))
            plt.show()

    def to_excel(self, name):
        # self.result.to_excel('bt统计.xlsx')
        # self.fund_pos_df.to_excel('bt权重.xlsx')
        # self.fund_value_df.to_excel('bt价值.xlsx')
        # self.data.to_excel('bt数据.xlsx')
        # self.fund_contribution_df.to_excel('bt贡献.xlsx')

        writer = pd.ExcelWriter(name + '.xlsx')
        self.result.to_excel(writer, sheet_name='统计')
        self.fund_pos_df.to_excel(writer, sheet_name='权重')
        self.fund_value_df.to_excel(writer, sheet_name='价值')
        self.data.to_excel(writer, sheet_name='数据')
        self.fund_contribution_df.to_excel(writer, sheet_name='贡献')
        writer.save()


def exposure_ts(nav_data, factor_data, freq, step='Forward'):
    exposure = pd.DataFrame()
    for i in range(freq + 1, len(nav_data) + 1):
        loadings = rp.loadings_matrix(
            X=factor_data.pct_change().iloc[i - freq: i, :],
            Y=nav_data.pct_change().iloc[i - freq: i, :],
            stepwise=step
        )
        loadings['t_date'] = nav_data.index[i - 1]
        exposure = exposure.append(loadings)
    return exposure










# -*- coding: utf-8 -*-

from hbshare.fe.asset_allocation.mean_variance_solver import MaxReturn, MinRisk
import numpy as np
import pandas as pd


class EF:
    def __init__(self, asset_type, asset_list, ret, cov, risk_free_rate, lb_list, ub_list, asset_category_df, total_weight, weight):
        self.asset_type = asset_type
        self.asset_list = asset_list
        self.ret = np.matrix(ret)
        self.cov = np.matrix(cov)
        self.target_risk = 99
        self.target_return = -99
        self.risk_free_rate = risk_free_rate
        self.lb_list = lb_list
        self.ub_list = ub_list
        self.asset_category_df = asset_category_df
        self.category_lb_list = [0.0, 0.0, 0.0, 0.0]
        self.category_ub_list = [1.0, 1.0, 1.0, 1.0]
        self.total_weight = total_weight
        self.weight = weight
        self.w_star = self.weight[self.asset_list].T if len(self.weight[self.asset_list].dropna()) > 0 else None

    def solve(self):
        # 最小方差组合（最左侧点）
        min_weight, min_status = MinRisk(self.asset_type, self.asset_list, self.ret, self.cov, self.target_return, self.lb_list, self.ub_list, self.asset_category_df, self.category_lb_list, self.category_ub_list, self.total_weight).solve()
        if min_status == 'optimal':
            min_weight = pd.DataFrame(min_weight).loc[self.asset_list]
            min_sigma2 = (np.matrix(min_weight).T * self.cov * np.matrix(min_weight))[0, 0]
            min_sigma = np.sqrt(min_sigma2)
            min_r = (np.matrix(min_weight).T * self.ret)[0, 0]
        else:
            min_weight = pd.DataFrame(index=self.asset_list, columns=[0], data=[np.nan] * len(self.asset_list))
            min_sigma = np.nan
            min_r = self.ret.mean()
        # 最大收益组合（最上侧点）
        max_weight, max_status = MaxReturn(self.asset_type, self.asset_list, self.ret, self.cov, self.target_risk, self.lb_list, self.ub_list, self.asset_category_df, self.category_lb_list, self.category_ub_list, self.total_weight).solve()
        if max_status == 'optimal':
            max_weight = pd.DataFrame(max_weight).loc[self.asset_list]
            max_sigma2 = (np.matrix(max_weight).T * self.cov * np.matrix(max_weight))[0, 0]
            max_sigma = np.sqrt(max_sigma2)
            max_r = (np.matrix(max_weight).T * self.ret)[0, 0]
        else:
            max_weight = pd.DataFrame(index=self.asset_list, columns=[0], data=[np.nan] * len(self.asset_list))
            max_sigma = np.nan
            max_r = self.ret.max()
        # 有效前沿曲线
        delta = ((max_r - min_r) / 20.0) - 1e-8
        ret_list = [(min_r + delta * i) for i in range(1, 20)]
        weight_P, sigma_P, r_P = [], [], []
        for ret in ret_list:
            weight, status = MinRisk(self.asset_type, self.asset_list, self.ret, self.cov, ret, self.lb_list, self.ub_list, self.asset_category_df, self.category_lb_list, self.category_ub_list, self.total_weight).solve()
            if status == 'optimal':
                weight = pd.DataFrame(weight).loc[self.asset_list]
            else:
                weight = pd.DataFrame(index=self.asset_list, columns=[0], data=[np.nan] * len(self.asset_list))
            sigma2 = (np.matrix(weight).T * self.cov * np.matrix(weight))[0, 0]
            sigma = np.sqrt(sigma2)
            r = (np.matrix(weight).T * self.ret)[0, 0]
            weight_P.append(weight)
            sigma_P.append(sigma)
            r_P.append(r)
        weight_P = [min_weight] + weight_P + [max_weight]
        sigma_P = [min_sigma] + sigma_P + [max_sigma]
        r_P = [min_r] + r_P + [max_r]
        # try:
        #     min_weight, min_status = MinRisk(self.asset_list, self.ret, self.cov, min_r, self.lb_list, self.ub_list, self.total_weight).solve()
        #     min_weight = pd.DataFrame(min_weight).loc[self.asset_list]
        #     min_sigma2 = (np.matrix(min_weight).T * self.cov * np.matrix(min_weight))[0, 0]
        #     min_sigma = np.sqrt(min_sigma2)
        #     min_r = (np.matrix(min_weight).T * self.ret)[0, 0]
        # except:
        #     pass
        # weight_P = [min_weight] + weight_P
        # sigma_P = [min_sigma] + sigma_P
        # r_P = [min_r] + r_P
        # try:
        #     max_weight, max_status = MinRisk(self.asset_list, self.ret, self.cov, max_r, self.lb_list, self.ub_list, self.total_weight).solve()
        #     max_weight = pd.DataFrame(max_weight).loc[self.asset_list]
        #     max_sigma2 = (np.matrix(max_weight).T * self.cov * np.matrix(max_weight))[0, 0]
        #     max_sigma = np.sqrt(max_sigma2)
        #     max_r = (np.matrix(max_weight).T * self.ret)[0, 0]
        # except:
        #     pass
        # weight_P = weight_P + [max_weight]
        # sigma_P = sigma_P + [max_sigma]
        # r_P = r_P + [max_r]
        # 优化组合
        if self.w_star is not None:
            sigma2_star = (np.matrix(self.w_star).T * self.cov * np.matrix(self.w_star))[0, 0]
            sigma_star = np.sqrt(sigma2_star)
            r_star = (np.matrix(self.w_star).T * self.ret)[0, 0]
        else:
            sigma_star = np.nan
            r_star = np.nan
        # # 有效前沿曲线绘制
        # from matplotlib.ticker import FuncFormatter
        # import matplotlib.pyplot as plt
        # plt.rcParams['font.sans-serif'] = ['SimHei']
        # plt.rcParams['axes.unicode_minus'] = False
        # def to_100percent_r2(temp, position):
        #     return '%.2f' % (temp * 100) + '%'
        # plt.figure(figsize=(6, 6))
        # plt.plot(sigma_P, r_P, 'o-', color='#F04950', label='有效前沿')
        # plt.plot(sigma_star, r_star, 'o', color='#6268A2', label='最优方案')
        # plt.gca().xaxis.set_major_formatter(FuncFormatter(to_100percent_r2))
        # plt.gca().yaxis.set_major_formatter(FuncFormatter(to_100percent_r2))
        # plt.xlabel('年化风险')
        # plt.ylabel('年化收益')
        # plt.title('{0}'.format(self.weight.index[0]))
        # plt.legend(loc=2)
        # plt.tight_layout()
        # plt.savefig('D:/Git/hbshare/hbshare/fe/xwq/data/aam/{0}.png'.format(self.weight.index[0]))

        # 数据结果
        ef_weight = pd.concat(weight_P, axis=1).T
        ef_weight.index = range(len(ef_weight))
        ef_weight.loc[:, 'cash'] = 1.0 - ef_weight.sum(axis=1)
        ef_weight.loc['result', :] = self.weight.iloc[0]
        sigma = pd.DataFrame(sigma_P)
        sigma.columns = ['sigma']
        sigma.loc['result', 'sigma'] = sigma_star
        r = pd.DataFrame(r_P)
        r.columns = ['r']
        r.loc['result', 'r'] = r_star
        ef_weight = pd.concat([ef_weight, sigma, r], axis=1)
        ef_weight = pd.concat([ef_weight.iloc[:-1].dropna(), ef_weight.iloc[-1:]])
        ef_weight = ef_weight.reset_index().rename(columns={'index': 'group'})
        return ef_weight


"""
大小盘流动性
"""
import pandas as pd
from datetime import datetime
import hbshare as hbs
from hbshare.fe.common.util.data_loader import get_trading_day_list
from WindPy import w

w.start()


class StockLiquidity:
    def __init__(self, start_date, end_date):
        self.start_date = start_date
        self.end_date = end_date
        self._load_data()

    def _load_data(self):
        # 指数
        week_list = get_trading_day_list(self.start_date, self.end_date, frequency="week")
        sql_script = "SELECT JYRQ as TRADEDATE, ZQDM, SPJG as TCLOSE from funddb.ZSJY WHERE " \
                     "ZQDM in ('000300', '000852', '399303') " \
                     "and JYRQ >= {} and JYRQ <= {}".format(self.start_date, self.end_date)
        data = pd.DataFrame(hbs.db_data_query('readonly', sql_script, page_size=10000)['data'])
        index_data = pd.pivot_table(
            data, index='TRADEDATE', columns='ZQDM', values='TCLOSE').reindex(week_list).dropna().sort_index()
        # DR007
        res = w.wsd("DR007.IB", "close", self.start_date, self.end_date, "")
        data = pd.DataFrame(res.Data[0], index=res.Times, columns=res.Codes)
        data.index.name = 'trade_date'
        data.reset_index(inplace=True)
        data['trade_date'] = data['trade_date'].apply(lambda x: datetime.strftime(x, '%Y%m%d'))
        data['DR007-MA20'] = data['DR007.IB'].rolling(20).mean()
        rates = data.set_index('trade_date').reindex(week_list)
        # 10年期国债收益率
        res = w.edb('M1001654', self.start_date, self.end_date)
        data = pd.DataFrame(res.Data[0], index=res.Times, columns=res.Codes)
        data.index.name = 'trade_date'
        data.reset_index(inplace=True)
        data['trade_date'] = data['trade_date'].apply(lambda x: datetime.strftime(x, '%Y%m%d'))
        data.rename(columns={"M1001654": "ytm_10y"}, inplace=True)
        ytm_data = data.set_index('trade_date').reindex(week_list)
        # 信用利差
        res = w.edb('M1016547,M1016548', self.start_date, self.end_date)
        data = pd.DataFrame(res.Data, index=res.Codes, columns=res.Times).T
        data.index.name = 'trade_date'
        data.reset_index(inplace=True)
        data['trade_date'] = data['trade_date'].apply(lambda x: datetime.strftime(x, '%Y%m%d'))
        data.rename(columns={"M1016547": "credit_spread_AA+", "M1016548": "credit_spread_AA"}, inplace=True)
        spread = data.set_index('trade_date').reindex(week_list)

        df = index_data.merge(
            rates, left_index=True, right_index=True).merge(
            ytm_data, left_index=True, right_index=True).merge(spread, left_index=True, right_index=True)

        df.rename(columns={"000300": "沪深300", "000852": "中证1000", "399303": "国证2000"}, inplace=True)
        df['中证1000/沪深300'] = df['中证1000'] / df['沪深300']
        df['国证2000/沪深300'] = df['国证2000'] / df['沪深300']

        self.data = df


if __name__ == '__main__':
    StockLiquidity('20141120', '20221118')


# -*- coding: utf-8 -*-

import pandas as pd
import pymysql
import hbshare as hbs
import warnings
warnings.filterwarnings('ignore', category=pymysql.Warning)


class Loader:
    def __init__(self):
        pass

    def get_df(self, sql, db, page_size=2000):
        data = hbs.db_data_query(db, sql, page_size=page_size, timeout=120)
        pages = data['pages']
        data = pd.DataFrame(data['data'])
        if pages > 1:
            for page in range(2, pages + 1):
                temp_data = hbs.db_data_query(db, sql, page_size=page_size, page_num=page, timeout=120)
                data = pd.concat([data, pd.DataFrame(temp_data['data'])], axis=0)
        return data

    def read_cal(self, start, end):
        sql = "SELECT jyrq, sfjj, sfzm, sfym FROM st_main.t_st_gg_jyrl WHERE jyrq>={0} and jyrq<={1}".format(start, end)
        df = self.get_df(sql, db='alluser')
        return df

    def read_mutual_index_daily_k_given_indexs(self, indexs, start, end):
        indexs = "'" + "','".join(indexs) + "'"
        sql = "SELECT zsdm AS INDEX_CODE, jzrq AS TRADE_DATE, spjg AS CLOSE_INDEX FROM st_fund.t_st_gm_clzs WHERE zsdm in ({0}) AND jzrq >= {1} AND jzrq <= {2} AND m_opt_type <> '03'".format(indexs, start, end)
        df = self.get_df(sql, db='funduser', page_size=200000)
        return df

    def read_private_index_daily_k_given_indexs(self, indexs, start, end):
        indexs = "'" + "','".join(indexs) + "'"
        sql = "SELECT zsdm AS INDEX_CODE, jyrq AS TRADE_DATE, spjg AS CLOSE_INDEX FROM st_hedge.t_st_sm_zhmzs WHERE zsdm in ({0}) AND jyrq >= {1} AND jyrq <= {2} AND m_opt_type <> '03'".format(indexs, start, end)
        df = self.get_df(sql, db='highuser', page_size=200000)
        return df

    def read_market_index_daily_k_given_index(self, index, start, end):
        df = hbs.commonQuery('MARKET_SPJG', zqdm=index, startDate=start, endDate=end, fields=['zqdm', 'jyrq', 'spjg'])
        df = df.rename(columns={'zqdm': 'INDEX_CODE', 'jyrq': 'TRADE_DATE', 'spjg': 'CLOSE_INDEX'})
        return df

    def read_mutual_fund_cumret_given_code(self, code, start, end):
        df = hbs.commonQuery('FUND_JJJZ', jjdm=code, startDate=start, endDate=end, fields=['jjdm', 'jzrq', 'fqdwjz'])
        df = df.rename(columns={'jjdm': 'FUND_CODE', 'jzrq': 'TRADE_DATE', 'fqdwjz': 'ADJ_NAV'})
        return df

    def read_private_fund_adj_nav_given_code(self, code, start, end):
        df = hbs.commonQuery('SIMU_JJJZ', jjdm=code, startDate=start, endDate=end, fields=['jjdm', 'jzrq', 'fqdwjz'])
        df = df.rename(columns={'jjdm': 'FUND_CODE', 'jzrq': 'TRADE_DATE', 'fqdwjz': 'ADJ_NAV'})
        return df

    def read_index_daily_k_given_indexs(self, mutual_indexs, private_indexs, market_indexs, start, end):
        df = hbs.commonQuery('MARKET_SPJG_BATCH', gmclzs=mutual_indexs, smclzs=private_indexs, sczs=market_indexs, startDate=start, endDate=end, fields=['zqdm', 'jyrq', 'spjg'])
        df = df.rename(columns={'zqdm': 'INDEX_CODE', 'jyrq': 'TRADE_DATE', 'spjg': 'CLOSE_INDEX'})
        return df

    def read_fund_cumret_given_codes(self, mutual_codes, private_codes, start, end):
        df = hbs.commonQuery('ALL_JJJZ', gmdm=mutual_codes, smdm=private_codes, startDate=start, endDate=end, fields=['jjdm', 'jzrq', 'fqdwjz'])
        df = df.rename(columns={'jjdm': 'FUND_CODE', 'jzrq': 'TRADE_DATE', 'fqdwjz': 'ADJ_NAV'})
        return df

    def read_market_index_market_value_given_indexs(self, indexs, start, end):
        indexs = "'" + "','".join(indexs) + "'"
        sql = "SELECT zqdm AS INDEX_CODE, jyrq AS TRADE_DATE, zsz AS MARKET_VALUE FROM st_market.t_st_zs_hqql WHERE zqdm in ({0}) AND jyrq >= {1} AND jyrq <= {2} AND m_opt_type <> '03'".format(indexs, start, end)
        df = self.get_df(sql, db='alluser', page_size=200000)
        return df

    def read_mutual_fund_market_value_given_codes(self, codes, start, end):
        codes = "'" + "','".join(codes) + "'"
        sql = "SELECT jjdm AS FUND_CODE, jsrq AS REPORT_DATE, jjzzc AS MARKET_VALUE FROM st_fund.t_st_gm_zcpz WHERE jjdm in ({0}) AND jsrq >= {1} AND jsrq <= {2} AND m_opt_type <> '03'".format(codes, start, end)
        df = self.get_df(sql, db='funduser', page_size=200000)
        return df


if __name__ == '__main__':
    import hbshare as hbs

    data = hbs.commonQuery('FUND_JJJZ',
                           jjdm='688888',
                           startDate='20230101', endDate='20230228')

    data = hbs.commonQuery('SIMU_JJJZ',
                       jjdm='SX8958',
                       startDate='20230101', endDate='20230228')

    data = hbs.commonQuery('MARKET_SPJG',
                       zqdm='000300',
                       startDate='20230101', endDate='20230228')

    print(data)


"""
风险控制专题报告-代码1，代码2在组合优化项目中
"""
import numpy as np
import pandas as pd
import hbshare as hbs
from tqdm import tqdm
import statsmodels.api as sm
import datetime
import os
from sqlalchemy import create_engine
from hbshare.quant.Kevin.rm_associated.config import engine_params
from hbshare.fe.common.util.config import style_name, industry_name
from hbshare.fe.common.util.data_loader import get_trading_day_list
from hbshare.quant.Kevin.rm_associated.config import style_names


def save_mkt_data_to_csv(start_date, end_date):
    trading_day_list = get_trading_day_list(start_date, end_date)

    for date in tqdm(trading_day_list):
        # MKT
        sql_script = "SELECT SYMBOL, TDATE, TCLOSE, VATURNOVER, PCHG, MCAP, TCAP FROM finchina.CHDQUOTE WHERE" \
                     " TDATE = {}".format(date)
        res = hbs.db_data_query('readonly', sql_script, page_size=6000)
        df = pd.DataFrame(res['data'])
        df = df[df['SYMBOL'].str[0].isin(['0', '3', '6'])]
        df.rename(columns={"SYMBOL": "ticker", "TDATE": "tradeDate", "TCAP": "marketValue",
                           "MCAP": "negMarketValue", "VATURNOVER": "turnoverValue",
                           "PCHG": "dailyReturnReinv"}, inplace=True)
        # BP
        sql_script = "SELECT a.TradingDay, a.PB, b.SecuCode From " \
                     "hsjy_gg.LC_DIndicesForValuation a join hsjy_gg.SecuMain b on a.InnerCode = b.InnerCode where " \
                     "to_char(a.TradingDay,'yyyymmdd') = {}".format(date)
        res = hbs.db_data_query('readonly', sql_script, page_size=6000)
        bp_data_main = pd.DataFrame(res['data']).rename(columns={"SECUCODE": "ticker"})[['ticker', 'PB']]
        sql_script = "SELECT a.TradingDay, a.PB, b.SecuCode From " \
                     "hsjy_gg.LC_STIBDIndiForValue a join hsjy_gg.SecuMain b on a.InnerCode = b.InnerCode where " \
                     "to_char(a.TradingDay,'yyyymmdd') = {}".format(date)
        res = hbs.db_data_query('readonly', sql_script, page_size=6000)
        if len(res['data']) == 0:
            bp_data = bp_data_main
        else:
            bp_data_stib = pd.DataFrame(res['data']).rename(columns={"SECUCODE": "ticker"})[['ticker', 'PB']]
            bp_data = pd.concat([bp_data_main, bp_data_stib])
            bp_data = bp_data[bp_data['ticker'].str[0].isin(['0', '3', '6'])]

        data = df.merge(bp_data, on='ticker')
        include_cols = ['ticker', 'negMarketValue', 'marketValue', 'turnoverValue', 'dailyReturnReinv', 'PB']
        data[include_cols].to_csv('D:\\MarketInfoSaver\\market_info_{}.csv'.format(date), index=False)
        print("{}: 有效数据{}条".format(date, data.shape[0]))


def cal_illiq_factor(start_date, end_date):
    """
    计算非流动性因子
    """
    month_list = get_trading_day_list(start_date, end_date, frequency="month")
    illiq = []
    for i in tqdm(range(1, len(month_list))):
        pre_date, t_date = month_list[i - 1], month_list[i]
        path = "D:\\MarketInfoSaver"
        listdir = os.listdir(path)
        listdir = [x for x in listdir if pre_date < x.split('_')[-1].split('.')[0] <= t_date]
        data = []
        for filename in listdir:
            trade_date = filename.split('.')[0].split('_')[-1]
            date_t_data = pd.read_csv(os.path.join(path, filename))
            date_t_data['ticker'] = date_t_data['ticker'].apply(lambda x: str(x).zfill(6))
            date_t_data['trade_date'] = trade_date
            data.append(date_t_data)
        data = pd.concat(data)
        data.loc[data['turnoverValue'] < 1e-8, 'dailyReturnReinv'] = np.NaN
        data['illiq'] = data['dailyReturnReinv'].abs() / (data['turnoverValue'] / 1e+9)
        data = pd.pivot_table(data, index='trade_date', columns='ticker', values='illiq').sort_index()

        na_counts = data.isnull().sum()
        exclude_list = na_counts[na_counts >= data.shape[0] * 0.7].index.tolist()

        t_factor = data.mean().to_frame('illiq')
        t_factor.loc[exclude_list] = np.NaN
        t_factor['trade_date'] = t_date
        t_factor = t_factor.reset_index()

        illiq.append(t_factor)

    illiq = pd.concat(illiq)
    illiq.to_csv("D:\\研究基地\\G-专题报告\\【2022.12】alpha策略的风控管理\\非流动性因子.csv", index=False)


def ols(y, x):
    x_ = sm.add_constant(x.copy())
    model = sm.OLS(y, x_, )
    results = model.fit()
    return np.std(results.resid), results.rsquared


def call_ivff_factor(start_date, end_date):
    """
    计算特质波动率（特异度）因子
    """
    month_list = get_trading_day_list(start_date, end_date, frequency="month")
    ivr_list = []
    for i in tqdm(range(1, len(month_list))):
        pre_date, t_date = month_list[i - 1], month_list[i]
        path = "D:\\MarketInfoSaver"
        listdir = os.listdir(path)
        listdir = [x for x in listdir if pre_date < x.split('_')[-1].split('.')[0] <= t_date]
        data = []
        for filename in listdir:
            trade_date = filename.split('.')[0].split('_')[-1]
            date_t_data = pd.read_csv(os.path.join(path, filename))
            date_t_data['ticker'] = date_t_data['ticker'].apply(lambda x: str(x).zfill(6))
            date_t_data['trade_date'] = trade_date
            data.append(date_t_data)
        data = pd.concat(data)
        data.loc[data['turnoverValue'] < 1e-8, 'dailyReturnReinv'] = np.NaN

        stock_return = data[['trade_date', 'ticker', 'dailyReturnReinv', 'negMarketValue', 'PB']].dropna()
        stock_return['negMarketValue'] /= 1e+9
        stock_return['dailyReturnReinv'] /= 100.

        grouped_df = stock_return.groupby('trade_date')
        quantile_df = grouped_df['negMarketValue'].quantile(0.33).to_frame('mkv_lower').merge(
            grouped_df['negMarketValue'].quantile(0.66).to_frame('mkv_upper'), left_index=True, right_index=True).merge(
            grouped_df['PB'].quantile(0.33).to_frame('pb_lower'), left_index=True, right_index=True).merge(
            grouped_df['PB'].quantile(0.66).to_frame('pb_upper'), left_index=True, right_index=True).reset_index()
        stock_return = stock_return.merge(quantile_df, on='trade_date')
        stock_return['w_ret'] = stock_return['dailyReturnReinv'] * stock_return['negMarketValue']
        # 市场收益
        market_return = stock_return.groupby('trade_date').apply(
            lambda x: (x['w_ret'] / x['negMarketValue'].sum()).sum())
        # 市值因子收益
        size_return_small = stock_return.groupby('trade_date').apply(
            lambda x: (x[x['negMarketValue'] <= x['mkv_lower']]['w_ret']).sum() /
                      (x[x['negMarketValue'] <= x['mkv_lower']]['negMarketValue']).sum())
        size_return_large = stock_return.groupby('trade_date').apply(
            lambda x: (x[x['negMarketValue'] >= x['mkv_upper']]['w_ret']).sum() /
                      (x[x['negMarketValue'] >= x['mkv_upper']]['negMarketValue']).sum())
        size_return = size_return_small - size_return_large
        # 估值因子收益
        pb_return_low = stock_return.groupby('trade_date').apply(
            lambda x: (x[x['PB'] <= x['pb_lower']]['w_ret']).sum() /
            (x[x['PB'] <= x['pb_lower']]['negMarketValue']).sum())
        pb_return_high = stock_return.groupby('trade_date').apply(
            lambda x: (x[x['PB'] >= x['pb_upper']]['w_ret']).sum() /
            (x[x['PB'] >= x['pb_upper']]['negMarketValue']).sum())
        pb_return = pb_return_low - pb_return_high
        reg_data = market_return.to_frame('beta').merge(
            size_return.to_frame('size'), left_index=True, right_index=True).merge(
            pb_return.to_frame('value'), left_index=True, right_index=True)
        reg_data = stock_return[['trade_date', 'ticker', 'dailyReturnReinv']].merge(reg_data, on='trade_date')
        # regression
        res = reg_data.groupby(by='ticker').apply(lambda x: ols(x['dailyReturnReinv'], x[['beta', 'size', 'value']]))
        index = [i[0] for i in reg_data.groupby(by='ticker')]

        ivol = pd.Series(index=index, data=[x[0] for x in res]) * np.sqrt(252)
        ivr = 1 - pd.Series(index=index, data=[x[1] for x in res])

        t_factor = ivol.to_frame('ivff').merge(ivr.to_frame('ivr'), left_index=True, right_index=True).reset_index()
        t_factor['trade_date'] = t_date

        ivr_list.append(t_factor)

    ivr_df = pd.concat(ivr_list)
    ivr_df.to_csv("D:\\研究基地\\G-专题报告\\【2022.12】alpha策略的风控管理\\特质波动率因子.csv", index=False)


def factor_preprocess(factor_df, factor_name):
    df = factor_df.copy()
    date_list = sorted(factor_df['trade_date'].unique())
    processed_factor = []
    for date in tqdm(date_list):
        t_df = df[df['trade_date'] == date].dropna()
        # 去极值
        median = df[factor_name].median()
        new_median = abs(df[factor_name] - median).median()
        up = median + 5 * new_median
        down = median - 5 * new_median
        t_df[factor_name] = t_df[factor_name].clip(down, up)
        # 中性化
        sql_script = "SELECT * FROM st_ashare.r_st_barra_style_factor where TRADE_DATE = '{}'".format(date)
        res = hbs.db_data_query('alluser', sql_script, page_size=5000)
        style_factor = pd.DataFrame(res['data'])[['ticker'] + style_names]
        t_df = t_df.merge(style_factor, on='ticker')
        x = np.array(t_df[style_names])
        x_ = sm.add_constant(x.copy())
        model = sm.OLS(t_df[factor_name], x_, )
        t_df[factor_name + '_neu'] = model.fit().resid
        # 标准化
        t_df[factor_name + '_std'] = \
            (t_df[factor_name + '_neu'] - t_df[factor_name + '_neu'].mean()) / t_df[factor_name + '_neu'].std()

        processed_factor.append(t_df)

    processed_factor = pd.concat(processed_factor)
    # processed_factor.groupby('trade_date').apply(lambda x: x.corr().loc['illiq', style_names]).mean()

    return processed_factor[['trade_date', 'ticker', factor_name, factor_name + '_neu', factor_name + '_std']]


def cal_factor_ic(factor_df, factor_name):
    df = factor_df.copy()
    month_list = get_trading_day_list('20191220', '20221120', frequency="month")
    ic_list = []
    for i in tqdm(range(1, len(month_list))):
        pre_date, t_date = month_list[i - 1], month_list[i]
        path = "D:\\MarketInfoSaver"
        listdir = os.listdir(path)
        listdir = [x for x in listdir if pre_date < x.split('_')[-1].split('.')[0] <= t_date]
        data = []
        for filename in listdir:
            trade_date = filename.split('.')[0].split('_')[-1]
            date_t_data = pd.read_csv(os.path.join(path, filename))
            date_t_data['ticker'] = date_t_data['ticker'].apply(lambda x: str(x).zfill(6))
            date_t_data['trade_date'] = trade_date
            data.append(date_t_data)
        data = pd.concat(data)
        data.loc[data['turnoverValue'] < 1e-8, 'dailyReturnReinv'] = np.NaN
        data['dailyReturnReinv'] /= 100.

        data = data.pivot(index='trade_date', columns='ticker', values='dailyReturnReinv').sort_index()
        future_ret = (1 + data).prod() - 1
        t_df = df[df['trade_date'] == pre_date].copy()
        t_df = t_df.set_index('ticker').merge(future_ret.to_frame('f_ret'), left_index=True, right_index=True)
        del t_df['trade_date']
        corr_df = t_df.corr().loc['f_ret', [factor_name, factor_name + '_neu', factor_name + '_std']].to_frame(pre_date)
        ic_list.append(corr_df)

    ic_df = pd.concat(ic_list, axis=1)
    print(ic_df.mean(axis=1))

    return ic_df

class PortfolioRiskPredict:
    """
    组合风险预测类
    """
    def __init__(self, trade_date, fund_name, fund_id, benchmark_id="000905"):
        self.trade_date = trade_date
        self.fund_name = fund_name
        self.fund_id = fund_id
        self.benchmark_id = benchmark_id
        self._load_data()

    def _load_shift_date(self):
        trade_dt = datetime.datetime.strptime(self.trade_date, '%Y%m%d')
        pre_date = (trade_dt - datetime.timedelta(days=100)).strftime('%Y%m%d')

        sql_script = "SELECT JYRQ, SFJJ, SFZM, SFYM FROM funddb.JYRL WHERE JYRQ >= {} and JYRQ <= {}".format(
            pre_date, self.trade_date)
        res = hbs.db_data_query('readonly', sql_script, page_size=5000)
        df = pd.DataFrame(res['data']).rename(
            columns={"JYRQ": 'calendarDate', "SFJJ": 'isOpen',
                     "SFZM": "isWeekEnd", "SFYM": "isMonthEnd"}).sort_values(by='calendarDate')
        df['isOpen'] = df['isOpen'].astype(int).replace({0: 1, 1: 0})
        df['isWeekEnd'] = df['isWeekEnd'].fillna(0).astype(int)
        df['isMonthEnd'] = df['isMonthEnd'].fillna(0).astype(int)

        trading_day_list = df[df['isMonthEnd'] == 1]['calendarDate'].tolist()

        return trading_day_list[-1]

    def _load_portfolio_weight_series(self):
        sql_script = "SELECT * FROM private_fund_holding where fund_name = '{}' and trade_date = {}".format(
            self.fund_name, self.trade_date)
        engine = create_engine(engine_params)
        holding_df = pd.read_sql(sql_script, engine)
        holding_df['trade_date'] = holding_df['trade_date'].apply(lambda x: datetime.datetime.strftime(x, '%Y%m%d'))

        return holding_df.set_index('ticker')['weight'] / 100.

    def _load_benchmark_weight_series(self, date):
        sql_script = "SELECT * FROM hsjy_gg.SecuMain where SecuCategory = 4 and SecuCode = '{}'".format(
            self.benchmark_id)
        res = hbs.db_data_query('readonly', sql_script)
        index_info = pd.DataFrame(res['data'])
        inner_code = index_info.set_index('SECUCODE').loc[self.benchmark_id, 'INNERCODE']

        sql_script = "SELECT (select a.SecuCode from hsjy_gg.SecuMain a where a.InnerCode = b.InnerCode and " \
                     "rownum = 1) SecuCode, b.EndDate, b.Weight FROM hsjy_gg.LC_IndexComponentsWeight b WHERE " \
                     "b.IndexCode = '{}' and b.EndDate = to_date('{}', 'yyyymmdd')".format(inner_code, date)
        data = pd.DataFrame(hbs.db_data_query('readonly', sql_script)['data'])
        weight_df = data.rename(
            columns={"SECUCODE": "consTickerSymbol", "ENDDATE": "effDate", "WEIGHT": "weight"})

        return weight_df.set_index('consTickerSymbol')['weight'] / 100.

    @staticmethod
    def _load_style_exposure(date):
        sql_script = "SELECT * FROM st_ashare.r_st_barra_style_factor where TRADE_DATE = '{}'".format(date)
        res = hbs.db_data_query('alluser', sql_script, page_size=5000)
        exposure_df = pd.DataFrame(res['data']).set_index('ticker')
        ind_names = [x.lower() for x in industry_name['sw'].values()]
        exposure_df = exposure_df[style_name + ind_names + ['country']]

        return exposure_df

    @staticmethod
    def _load_factor_cov(date):
        sql_script = "SELECT * FROM st_ashare.r_st_barra_factor_cov where TRADE_DATE = '{}'".format(date)
        res = hbs.db_data_query('alluser', sql_script, page_size=1000)
        factor_cov = pd.DataFrame(res['data'])
        factor_cov['factor_name'] = factor_cov['factor_name'].apply(lambda x: x.lower())
        ind_names = [x.lower() for x in industry_name['sw'].values()]
        factor_list = style_name + ind_names + ['country']
        factor_cov = factor_cov.set_index('factor_name').reindex(factor_list)[factor_list]

        return factor_cov

    @staticmethod
    def _load_srisk(date):
        sql_script = "SELECT * FROM st_ashare.r_st_barra_s_risk where TRADE_DATE = '{}'".format(date)
        res = hbs.db_data_query('alluser', sql_script, page_size=5000)
        srisk = pd.DataFrame(res['data']).set_index('ticker')
        srisk.rename(columns={"s_ret": "srisk"}, inplace=True)

        return srisk[['srisk']]

    def _load_fund_ret(self, start_date):
        sql_params = {
            "ip": "192.168.223.152",
            "user": "readonly",
            "pass": "c24mg2e6",
            "port": "3306",
            "database": "work"
        }
        engine_params1 = "mysql+pymysql://{}:{}@{}:{}/{}".format(sql_params['user'], sql_params['pass'],
                                                                 sql_params['ip'],
                                                                 sql_params['port'], sql_params['database'])
        engine1 = create_engine(engine_params1)
        sql_script = "SELECT * FROM daily_nav where code = '{}' and date >= '{}'".format(
            self.fund_id, start_date)
        data = pd.read_sql(sql_script, engine1)
        data['date'] = data['date'].map(str)

        nav_df = data.set_index('date')['fqjz']

        sql_script = "SELECT JYRQ as TRADEDATE, ZQMC as INDEXNAME, SPJG as TCLOSE from funddb.ZSJY WHERE ZQDM = '{}' " \
                     "and JYRQ >= {} and JYRQ <= {}".format(self.benchmark_id, nav_df.index[0], nav_df.index[-1])
        index_data = pd.DataFrame(
            hbs.db_data_query('readonly', sql_script, page_size=5000)['data']).set_index('TRADEDATE')['TCLOSE']
        index_nav = index_data.reindex(nav_df.index).to_frame('benchmark')

        assert (nav_df.shape[0] == index_nav.shape[0])

        excess_return = nav_df.pct_change().sub(index_nav.pct_change()['benchmark'].squeeze(), axis=0)
        excess_return = excess_return.dropna()

        return excess_return.head(21)

    def _load_data(self):
        shift_date = self._load_shift_date()
        self.portfolio_weight_series = self._load_portfolio_weight_series()
        self.benchmark_weight_series = self._load_benchmark_weight_series(shift_date)
        self.exposure_df = self._load_style_exposure(shift_date)
        self.factor_cov = self._load_factor_cov(shift_date)
        self.srisk = self._load_srisk(shift_date)
        self.fund_ret = self._load_fund_ret(shift_date)

    def run(self):
        weight_df = self.portfolio_weight_series.to_frame('port').merge(
            self.benchmark_weight_series.to_frame('bm'), left_index=True, right_index=True, how='outer').fillna(0.)
        weight_df['active'] = weight_df['port'] - weight_df['bm']
        exposure_df = self.exposure_df.astype(float)
        factor_cov = self.factor_cov.divide(12 * 10000)
        srisk = self.srisk ** 2 * 21

        idx = set(weight_df.index).intersection(set(exposure_df.index)).intersection(set(srisk.index))

        weight = weight_df.reindex(idx)[['active']]
        exposure_df = exposure_df.reindex(idx)
        srisk = srisk.reindex(idx)['srisk']

        common_risk = np.mat(weight).T.dot(
            np.mat(exposure_df).dot(np.mat(factor_cov)).dot(np.mat(exposure_df).T)).dot(np.mat(weight))
        specific_risk = np.mat(weight).T.dot(np.diag(srisk)).dot(np.mat(weight))

        risk = np.sqrt(np.array(common_risk + specific_risk)[0][0]) * np.sqrt(12)
        # common_risk = np.array(common_risk)[0][0]
        # specific_risk = np.array(specific_risk)[0][0]

        actual_risk = self.fund_ret.std() * np.sqrt(252)

        print("{}_{}：预期风险: {}, 实际风险: {}".format(self.fund_name, self.trade_date, risk, actual_risk))

        # a = pd.Series(data=[common_risk * 12, specific_risk * 12, risk, actual_risk])

        return risk, actual_risk


if __name__ == '__main__':
    save_mkt_data_to_csv('20230117', '20230117')
    # cal_illiq_factor('20191120', '20221130')
    # call_ivff_factor('20191120', '20221130')

    # factor = pd.read_csv("D:\\研究基地\\G-专题报告\\【2022.12】alpha策略的风控管理\\特质波动率因子.csv", dtype={"trade_date": str})
    # factor.rename(columns={"index": "ticker"}, inplace=True)
    # factor['ticker'] = factor['ticker'].apply(lambda x: str(x).zfill(6))
    # p_factor = factor_preprocess(factor, 'ivr')
    # ic = cal_factor_ic(p_factor, 'ivr')
    # p_factor.to_csv('D:\\研究基地\\G-专题报告\\【2022.12】alpha策略的风控管理\\特质波动率因子_处理后.csv', index=False)

    # factor = pd.read_csv("D:\\研究基地\\G-专题报告\\alpha策略的风控管理\\非流动性因子.csv", dtype={"trade_date": str})
    # factor['ticker'] = factor['ticker'].apply(lambda x: str(x).zfill(6))
    # p_factor = factor_preprocess(factor, 'illiq')
    # ic = cal_factor_ic(p_factor, 'illiq')

    # ic_df = pd.read_excel("D:\\研究基地\\G-专题报告\\alpha策略的风控管理\\数据记录.xlsx", sheet_name=1, index_col=0)

    # PortfolioRiskPredict('20220630', '星阔广厦1号中证500指数增强', 'SNU706').run()
    # PortfolioRiskPredict('20220630', '因诺聚配中证500指数增强', 'SGX346').run()
    # PortfolioRiskPredict('20220630', '赫富500指数增强一号', 'SEP463').run()
    # PortfolioRiskPredict('20220630', '概率500指增2号', 'STJ128').run()
    # PortfolioRiskPredict('20220630', '伯兄熙宁', 'STR792').run()
    # PortfolioRiskPredict('20220630', '白鹭精选量化鲲鹏十号', 'P48487').run()
    # PortfolioRiskPredict('20220630', '乾象中证500指数增强1号', 'P48470').run()


# -*- coding: utf-8 -*-

from scipy.optimize import minimize
import cvxpy as cp
import numpy as np
import pandas as pd


class MaxReturn:
    def __init__(self, asset_type, asset_list, ret, cov, target_risk, lb_list, ub_list, asset_category_df, category_lb_list, category_ub_list, total_weight):
        self.asset_type = asset_type
        self.asset_list = asset_list
        self.ret = np.matrix(ret)
        self.cov = np.matrix(cov)
        self.target_risk = target_risk
        self.lb_list = lb_list
        self.ub_list = ub_list
        self.total_weight = total_weight
        if self.asset_type == 'fund':
            self.asset_category = asset_category_df.reset_index()
            self.asset_category['IS_CATEGORY'] = 1
            self.asset_category = self.asset_category.pivot(index='CATEGORY', columns='index', values='IS_CATEGORY')
            self.asset_category = self.asset_category[self.asset_list].reindex(['EQUITY', 'BOND', 'COMMODITY', 'CURRENCY']).fillna(0)
            self.asset_category = np.matrix(self.asset_category)
            self.category_lb_list = category_lb_list
            self.category_ub_list = category_ub_list

    def solve(self):
        n = len(self.asset_list)
        w = cp.Variable(n)
        ret = self.ret.T @ w
        risk = cp.quad_form(w, self.cov)
        if self.asset_type == 'fund':
            category_cons = self.asset_category @ w
            prob = cp.Problem(cp.Maximize(ret),
                              [risk <= self.target_risk ** 2.0,
                               cp.sum(w) == self.total_weight,
                               w >= self.lb_list,
                               w <= self.ub_list,
                               category_cons >= self.category_lb_list,
                               category_cons <= self.category_ub_list])
        else:
            prob = cp.Problem(cp.Maximize(ret),
                              [risk <= self.target_risk ** 2.0,
                               cp.sum(w) == self.total_weight,
                               w >= self.lb_list,
                               w <= self.ub_list])
        prob.solve()
        if prob.status != 'optimal':
            weight = pd.Series(index=self.asset_list + ['cash'], data=[np.nan] * (len(self.asset_list) + 1))
            status = 'infeasible'
        else:
            arr_x = w.value
            for i in range(len(self.asset_list)):
                if arr_x[i] < self.lb_list[i]:
                    arr_x[i] = self.lb_list[i]
            for i in range(len(self.asset_list)):
                if arr_x[i] > self.ub_list[i]:
                    arr_x[i] = self.ub_list[i]
            if sum(arr_x) != self.total_weight:
                arr_x = arr_x / arr_x.sum() * self.total_weight

            weight = pd.Series(index=self.asset_list + ['cash'], data=list(arr_x) + [1.0 - arr_x.sum()])
            status = 'optimal'
        return weight, status


class MinRisk:
    def __init__(self, asset_type, asset_list, ret, cov, target_return, lb_list, ub_list, asset_category_df, category_lb_list, category_ub_list, total_weight):
        self.asset_type = asset_type
        self.asset_list = asset_list
        self.ret = np.matrix(ret)
        self.cov = np.matrix(cov)
        self.target_return = target_return
        self.lb_list = lb_list
        self.ub_list = ub_list
        self.total_weight = total_weight
        if self.asset_type == 'fund':
            self.asset_category = asset_category_df.reset_index()
            self.asset_category['IS_CATEGORY'] = 1
            self.asset_category = self.asset_category.pivot(index='CATEGORY', columns='index', values='IS_CATEGORY')
            self.asset_category = self.asset_category[self.asset_list].reindex(['EQUITY', 'BOND', 'COMMODITY', 'CURRENCY']).fillna(0)
            self.asset_category = np.matrix(self.asset_category)
            self.category_lb_list = category_lb_list
            self.category_ub_list = category_ub_list

    def solve(self):
        n = len(self.asset_list)
        w = cp.Variable(n)
        ret = self.ret.T @ w
        risk = cp.quad_form(w, self.cov)
        if self.asset_type == 'fund':
            category_cons = self.asset_category @ w
            prob = cp.Problem(cp.Minimize(risk),
                              [ret >= self.target_return,
                               cp.sum(w) == self.total_weight,
                               w >= self.lb_list,
                               w <= self.ub_list,
                               category_cons >= self.category_lb_list,
                               category_cons <= self.category_ub_list])
        else:
            prob = cp.Problem(cp.Minimize(risk),
                              [ret >= self.target_return,
                               cp.sum(w) == self.total_weight,
                               w >= self.lb_list,
                               w <= self.ub_list])
        prob.solve()
        if prob.status != 'optimal':
            weight = pd.Series(index=self.asset_list + ['cash'], data=[np.nan] * (len(self.asset_list) + 1))
            status = 'infeasible'
        else:
            arr_x = w.value
            for i in range(len(self.asset_list)):
                if arr_x[i] < self.lb_list[i]:
                    arr_x[i] = self.lb_list[i]
            for i in range(len(self.asset_list)):
                if arr_x[i] > self.ub_list[i]:
                    arr_x[i] = self.ub_list[i]
            if sum(arr_x) != self.total_weight:
                arr_x = arr_x / arr_x.sum() * self.total_weight

            weight = pd.Series(index=self.asset_list + ['cash'], data=list(arr_x) + [1.0 - arr_x.sum()])
            status = 'optimal'
        return weight, status


class MaxUtility:
    def __init__(self, asset_type, asset_list, ret, cov, risk_aversion, lb_list, ub_list, asset_category_df, category_lb_list, category_ub_list, total_weight):
        self.asset_type = asset_type
        self.asset_list = asset_list
        self.ret = np.matrix(ret)
        self.cov = np.matrix(cov)
        self.risk_aversion = risk_aversion
        self.lb_list = lb_list
        self.ub_list = ub_list
        self.total_weight = total_weight
        if self.asset_type == 'fund':
            self.asset_category = asset_category_df.reset_index()
            self.asset_category['IS_CATEGORY'] = 1
            self.asset_category = self.asset_category.pivot(index='CATEGORY', columns='index', values='IS_CATEGORY')
            self.asset_category = self.asset_category[self.asset_list].reindex(['EQUITY', 'BOND', 'COMMODITY', 'CURRENCY']).fillna(0)
            self.asset_category = np.matrix(self.asset_category)
            self.category_lb_list = category_lb_list
            self.category_ub_list = category_ub_list

    def solve(self):
        n = len(self.asset_list)
        w = cp.Variable(n)
        ret = self.ret.T @ w
        risk = cp.quad_form(w, self.cov)
        if self.asset_type == 'fund':
            category_cons = self.asset_category @ w
            prob = cp.Problem(cp.Maximize(ret - 0.5 * self.risk_aversion * risk),
                              [cp.sum(w) == self.total_weight,
                               w >= self.lb_list,
                               w <= self.ub_list,
                               category_cons >= self.category_lb_list,
                               category_cons <= self.category_ub_list])
        else:
            prob = cp.Problem(cp.Maximize(ret - 0.5 * self.risk_aversion * risk),
                              [cp.sum(w) == self.total_weight,
                               w >= self.lb_list,
                               w <= self.ub_list])
        prob.solve()
        if prob.status != 'optimal':
            weight = pd.Series(index=self.asset_list + ['cash'], data=[np.nan] * (len(self.asset_list) + 1))
            status = 'infeasible'
        else:
            arr_x = w.value
            for i in range(len(self.asset_list)):
                if arr_x[i] < self.lb_list[i]:
                    arr_x[i] = self.lb_list[i]
            for i in range(len(self.asset_list)):
                if arr_x[i] > self.ub_list[i]:
                    arr_x[i] = self.ub_list[i]
            if sum(arr_x) != self.total_weight:
                arr_x = arr_x / arr_x.sum() * self.total_weight

            weight = pd.Series(index=self.asset_list + ['cash'], data=list(arr_x) + [1.0 - arr_x.sum()])
            status = 'optimal'
        return weight, status


class MSR:
    def __init__(self, asset_type, asset_list, ret, cov, lb_list, ub_list, asset_category_df, category_lb_list, category_ub_list, total_weight):
        self.asset_type = asset_type
        self.asset_list = asset_list
        self.ret = np.matrix(ret)
        self.cov = np.matrix(cov)
        self.lb_list = lb_list
        self.ub_list = ub_list
        self.total_weight = total_weight
        if self.asset_type == 'fund':
            self.asset_category = asset_category_df.reset_index()
            self.asset_category['IS_CATEGORY'] = 1
            self.asset_category = self.asset_category.pivot(index='CATEGORY', columns='index', values='IS_CATEGORY')
            self.asset_category = self.asset_category[self.asset_list].reindex(['EQUITY', 'BOND', 'COMMODITY', 'CURRENCY']).fillna(0)
            self.asset_category = np.matrix(self.asset_category)
            self.category_lb_list = category_lb_list
            self.category_ub_list = category_ub_list

    def solve(self):
        n = len(self.asset_list)
        w = cp.Variable(n)
        ret = self.ret.T @ w
        risk = cp.quad_form(w, self.cov)
        if self.asset_type == 'fund':
            category_cons = self.asset_category @ w
            prob = cp.Problem(cp.Minimize(risk),
                              [ret == self.ret.mean(),
                               w >= self.lb_list * cp.sum(w),
                               w <= self.ub_list * cp.sum(w),
                               category_cons >= self.category_lb_list * cp.sum(w),
                               category_cons <= self.category_ub_list * cp.sum(w)])
        else:
            prob = cp.Problem(cp.Minimize(risk),
                              [ret == self.ret.mean(),
                               w >= self.lb_list * cp.sum(w),
                               w <= self.ub_list * cp.sum(w)])
        prob.solve()
        if prob.status != 'optimal':
            weight = pd.Series(index=self.asset_list + ['cash'], data=[np.nan] * (len(self.asset_list) + 1))
            status = 'infeasible'
        else:
            arr_x = w.value
            arr_x = arr_x / arr_x.sum() * self.total_weight
            for i in range(len(self.asset_list)):
                if arr_x[i] < self.lb_list[i]:
                    arr_x[i] = self.lb_list[i]
            for i in range(len(self.asset_list)):
                if arr_x[i] > self.ub_list[i]:
                    arr_x[i] = self.ub_list[i]
            if sum(arr_x) != self.total_weight:
                arr_x = arr_x / arr_x.sum() * self.total_weight

            weight = pd.Series(index=self.asset_list + ['cash'], data=list(arr_x) + [1.0 - arr_x.sum()])
            status = 'optimal'
        return weight, status


# class MSR:
#     def __init__(self, asset_list, ret, cov, lb_list, ub_list, total_weight):
#         self.asset_list = asset_list
#         self.ret = np.matrix(ret)
#         self.cov = np.matrix(cov)
#         self.lb_list = lb_list
#         self.ub_list = ub_list
#         self.total_weight = total_weight
#
#     def _target_func(self, x):
#         x = np.matrix(x).T
#         ret = float(x.T * self.ret)
#         std = float(np.sqrt(x.T * self.cov * x))
#         sharpe = - ret / std
#         return sharpe
#
#     def _get_cons(self):
#         cons = ({'type': 'eq', 'fun': lambda x: sum(x) - self.total_weight})
#         return cons
#
#     def _get_bnds(self):
#         x0 = 1 / (np.diag(self.cov) ** 0.5)
#         x0 = x0 / sum(x0)
#         bnds = tuple((self.lb_list[i], self.ub_list[i]) for i in range(len(x0)))
#         return bnds, x0
#
#     def solve(self, ftol=1e-20):
#         cons = self._get_cons()
#         bnds, x0 = self._get_bnds()
#         options = {'disp': False, 'maxiter': 500, 'ftol': ftol}
#
#         res = minimize(self._target_func, x0, bounds=bnds, constraints=cons, method='SLSQP', options=options)
#         if not res.success:
#             weight = pd.Series(index=self.asset_list + ['cash'], data=[np.nan] * (len(self.asset_list) + 1))
#             status = 'infeasible'
#         else:
#             arr_x = np.array(list(res.x))
#             for i in range(len(self.asset_list)):
#                 if arr_x[i] < self.lb_list[i]:
#                     arr_x[i] = self.lb_list[i]
#             for i in range(len(self.asset_list)):
#                 if arr_x[i] > self.ub_list[i]:
#                     arr_x[i] = self.ub_list[i]
#             if sum(arr_x) != self.total_weight:
#                 arr_x = arr_x / arr_x.sum() * self.total_weight
#
#             weight = pd.Series(index=self.asset_list + ['cash'], data=list(arr_x) + [1.0 - arr_x.sum()])
#             status = 'optimal'
#         return weight, status


"""
可投池多策略构建 vs 多策略
"""
import pandas as pd
import numpy as np
from datetime import datetime
import hbshare as hbs
from hbshare.quant.Kevin.quant_room.MyUtil.data_loader import get_fund_nav_from_sql, get_trading_day_list
from Arbitrage_backtest import cal_annual_return, cal_annual_volatility, cal_sharpe_ratio, cal_max_drawdown


def get_fund_nav(sub_pool, name):
    fund_dict = sub_pool.set_index('代表产品')['基金代码'].to_dict()
    start_date, end_date = sub_pool['入池时间'].min(), '20220930'

    trading_day_list = get_trading_day_list(start_date, end_date, frequency="week")
    nav_df = get_fund_nav_from_sql(start_date, end_date, fund_dict).reindex(trading_day_list)

    adj_nav = []
    for fund_name in fund_dict.keys():
        s_date = sub_pool[sub_pool['代表产品'] == fund_name]['入池时间'].values[0]
        e_date = sub_pool[sub_pool['代表产品'] == fund_name]['出池时间'].values[0]
        e_date = end_date if e_date >= end_date else e_date
        tmp = nav_df[[fund_name]].loc[s_date: e_date]
        adj_nav.append(tmp)
    adj_nav = pd.concat(adj_nav, axis=1).sort_index()
    return_df = adj_nav.pct_change(limit=1).dropna(how='all', axis=1)
    mean_df = return_df.mean(axis=1).dropna().to_frame(name)

    return mean_df


def multi_strategy_portfolio():
    # 存续池
    pool_exist = pd.read_excel('D:\\研究基地\\私募量化基金池\\私募量化基金池 - 202209.xlsx', sheet_name="量化池列表", header=1)
    include_cols = ['基金代码', '代表产品', '超一级/一级策略', '二级策略', '入池时间']
    pool_exist = pool_exist.dropna(subset=["二级策略"])[include_cols]
    pool_exist['入池时间'] = pool_exist['入池时间'].apply(lambda x: datetime.strftime(x, '%Y%m%d'))
    # 已出池
    pool_out = pd.read_excel('D:\\研究基地\\私募量化基金池\\私募量化基金池 - 202209.xlsx', sheet_name="出池记录", header=1)
    include_cols = ['基金代码', '基金简称', '超一级/一级策略', '二级策略', '入池时间', '出池时间']
    pool_out = pool_out.dropna(subset=["二级策略"])[include_cols].rename(columns={"基金简称": "代表产品"})
    pool_out['入池时间'] = pool_out['入池时间'].apply(lambda x: datetime.strftime(x, '%Y%m%d'))
    pool_out['出池时间'] = pool_out['出池时间'].apply(lambda x: datetime.strftime(x, '%Y%m%d'))
    # 整合
    pool = pd.concat([pool_exist, pool_out]).sort_values(by='基金代码')
    pool['出池时间'] = pool['出池时间'].fillna('20301231')

    # 500指增
    sub_pool = pool[pool['超一级/一级策略'] == '股票 - 指数增强']
    sub_pool = sub_pool[sub_pool['二级策略'] == '500指数增强']
    mean_df_alpha = get_fund_nav(sub_pool, '500指增')
    # CTA
    sub_pool = pool[pool['超一级/一级策略'] == '管理期货']
    sub_pool = sub_pool[sub_pool['二级策略'] != '趋势CTA（美元）']
    sub_pool.loc[sub_pool['代表产品'].str.contains('黑翼'), '出池时间'] = '20201231'
    exclude_list = ['道合东哥2号', '展弘稳进1号1期', '德劭锐哲中国', '腾胜中国聚量宏观策略1号', '涵德盈冲量化CTA2号']
    sub_pool = sub_pool[~sub_pool['代表产品'].isin(exclude_list)]
    mean_df_cta = get_fund_nav(sub_pool, 'CTA')
    # 套利
    sub_pool1 = pool[(pool['超一级/一级策略'] == '另类策略') | (pool['二级策略'] == '套利CTA')]
    sub_pool2 = pool[pool['代表产品'].isin(['蒙玺分形3号', '稳博中睿6号'])]
    sub_pool = pd.concat([sub_pool1, sub_pool2])
    mean_df_arb = get_fund_nav(sub_pool, '套利')
    # 市场中性
    sub_pool = pool[pool['二级策略'] == '500市场中性']
    exclude_list = ['希格斯招享1号', '量派睿核10号']
    sub_pool = sub_pool[~sub_pool['代表产品'].isin(exclude_list)]
    mean_df_neu = get_fund_nav(sub_pool, '市场中性')

    mean_df = mean_df_alpha.merge(mean_df_cta, left_index=True, right_index=True, how='outer').merge(
        mean_df_arb, left_index=True, right_index=True, how='outer').merge(
        mean_df_neu, left_index=True, right_index=True, how='outer').sort_index().loc["20190215":]

    # 组合构建
    weight = [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1], [0.25, 0.2, 0, 0.55],
              [0.5, 0.5, 0, 0], [0.2, 0.05, 0.05, 0.7], [0.8, 0.2, 0, 0], [0, 0.1, 0.1, 0.8], [1./3, 1./3, 0, 1./3]]
    weight_df = pd.DataFrame(index=['500指增', 'CTA', '套利', '市场中性',
                                    '25%指增+55%市场中性+20%CTA',
                                    '50%指增+50%CTA', '20%指增+70%中性+5%CTA+5%套利',
                                    '80%指增+20%CTA', '80%中性+10%CTA+10%套利', '指增/CTA/市场中性各1/3'], columns=mean_df.columns, data=weight)
    port_return = mean_df.dot(weight_df.T)
    port_nav = (1 + port_return).cumprod()

    portfolio_index_df = pd.DataFrame(
        index=port_nav.columns, columns=['年化收益', '年化波动', '最大回撤', 'Sharpe', '胜率', '平均损益比'])
    portfolio_index_df.loc[:, '年化收益'] = port_nav.pct_change().dropna(how='all').apply(cal_annual_return, axis=0)
    portfolio_index_df.loc[:, '年化波动'] = \
        port_nav.pct_change().dropna(how='all').apply(cal_annual_volatility, axis=0)
    portfolio_index_df.loc[:, '最大回撤'] = port_nav.apply(cal_max_drawdown, axis=0)
    portfolio_index_df.loc[:, 'Sharpe'] = \
        port_nav.pct_change().dropna(how='all').apply(lambda x: cal_sharpe_ratio(x, 0.015), axis=0)
    portfolio_index_df.loc[:, '胜率'] = \
        port_nav.pct_change().dropna(how='all').apply(lambda x: x.gt(0).sum() / len(x), axis=0)
    portfolio_index_df.loc[:, '平均损益比'] = \
        port_nav.pct_change().dropna(how='all').apply(lambda x: x[x > 0].mean() / x[x < 0].abs().mean(), axis=0)
    portfolio_index_df.index.name = '产品名称'
    portfolio_index_df.reset_index(inplace=True)
    # 格式处理
    portfolio_index_df['年化收益'] = portfolio_index_df['年化收益'].apply(lambda x: format(x, '.2%'))
    portfolio_index_df['年化波动'] = portfolio_index_df['年化波动'].apply(lambda x: format(x, '.2%'))
    portfolio_index_df['最大回撤'] = portfolio_index_df['最大回撤'].apply(lambda x: format(x, '.2%'))
    portfolio_index_df['Sharpe'] = portfolio_index_df['Sharpe'].round(2)
    portfolio_index_df['胜率'] = portfolio_index_df['胜率'].apply(lambda x: format(x, '.1%'))
    portfolio_index_df['平均损益比'] = portfolio_index_df['平均损益比'].round(2)

    return portfolio_index_df


if __name__ == '__main__':
    multi_strategy_portfolio()